from dawis.make_regions import *
from dawis.datacube import *
from dawis.wavelet_transforms import *
from dawis.hard_threshold import *
from dawis.label_regions import *
from dawis.anscombe_transforms import *
from dawis.make_regions import *
from dawis.make_interscale_trees import *
from dawis.restore_objects import *
from dawis.gif import *
from dawis.exceptions import *
from dawis.table_noise_stds import *
from dawis.synthesis_by_analysis import *
from dawis.test import *



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# This file is part of DAWIS.
#
# name : anscombe_transform.py
# author : amael ellien
# last modification : 07/2021
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# anscombe transform of an image.
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

import numpy as np
from astropy.stats import sigma_clip
import logging

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def pg_noise_bissection(image, max_err = 1E-6, n_sigmas = 3, verbose = False):
    '''
    doc tot do    '''

    noise_pts = np.copy(image)
    noise_pixels = sigma_clip(image, sigma = n_sigmas, sigma_lower = n_sigmas, masked = False) # the max parameters are given

    sigma_max = np.nanstd(noise_pixels)           # by the parameters of a gaussian
    mean_max = np.nanmean(noise_pixels)              # noise estimation on the input
    gain_max = np.nanmax(noise_pixels) - np.nanmin(noise_pixels)              # image.

    sigma_min = 0.0          # the min parameters are given
    mean_min = 0.0           # by 0.
    gain_min = 0.0           #

    error = 1.0    # initial parameters for the
    step = 0       # bissection-like method.

    if verbose == True:
        log = logging.getLogger(__name__)
        log.info('bissection-like noise estimation')

    while error > max_err:

        sigma = ( sigma_min + sigma_max ) / 2.0 # the noise parameters for this
        mean = ( mean_min + mean_max ) / 2.0    # iteration are the mean of max
        gain = ( gain_min + gain_max ) / 2.0    # and min noise parameters.

        sigma_ansc = np.nanstd(sigma_clip(anscombe_transform(noise_pts, sigma = sigma, mean = mean, gain = gain), \
                                        sigma = n_sigmas, sigma_lower = n_sigmas, masked = False))

        if sigma_ansc <= 1.0:  # gaussian distribution countepart
            sigma_max = sigma  # overestimated.
            gain_max = gain
            mean_max = mean
        else:
            sigma_min = sigma # gaussian distribution
            gain_min = gain   # underestimated.
            mean_min = mean

        error = sigma_max - sigma_min # convergence of the algorithm.
        step += 1

        if verbose == True:
            log.info('step %02d sigma_ansc %f sigma %f mean %f gain %f' %(step,sigma_ansc,sigma,mean,gain))
    if verbose == True:
        log.info('done.')

    return sigma, mean, gain


def anscombe_transform(image, sigma = 0.0, mean = 0.0, gain = 1.0):

    '''
    apply anscombe transform to input data. doc to do
    '''

    out_image = np.copy( image )
    out_image = ( 2.0 / gain ) * np.sqrt( gain * out_image + ( 3. / 8. ) \
                        * ( gain ** 2 ) + ( sigma ** 2 ) - gain * mean )
    out_image[ np.where( np.isnan( out_image ) == True ) ] = 0.0
    return out_image

if __name__ == '__main__':

    from atrous import atrous
    from numpy.random import normal
    from astropy.io import fits
    from wavelet_transforms import bspl_atrous, wavelet_datacube
    from hard_threshold import hard_threshold
    from label_regions import label_regions

    hdu = fits.open('/home/ellien/wavelets/A1365.rebin.fits')
    im = hdu[0].data
    header = hdu[0].header

    sigma, mean, gain = pg_noise_bissection(im, max_err = 1E-6, n_sigmas = 3)
    aim = anscombe_transform(im, sigma, mean, gain)
    acdc, awdc = bspl_atrous(aim, 10, header)
    sdc = hard_threshold(awdc, n_sigmas = 5)
    ldc = label_regions(sdc)

    cdc, wdc = bspl_atrous(im, 10, header)

    acdc.to_fits('/home/ellien/devd/tests/A1365.rebin.acdc.fits', overwrite = True)
    cdc.to_fits('/home/ellien/devd/tests/A1365.rebin.cdc.fits', overwrite = True)
    awdc.to_fits('/home/ellien/devd/tests/A1365.rebin.awdc.fits', overwrite = True)
    wdc.to_fits('/home/ellien/devd/tests/A1365.rebin.wdc.fits', overwrite = True)
    sdc.to_fits('/home/ellien/devd/tests/A1365.rebin.sdc.fits', overwrite = True)
    ldc.to_fits('/home/ellien/devd/tests/A1365.rebin.ldc.fits', overwrite = True)



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# This file is part of DAWIS.
#
# NAME : make regions.py
# AUTHOR : Amael Ellien
# LAST MODIFICATION : 07/2021
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

from dawis.table_noise_stds import table_noise_stds
from skimage.measure import regionprops
from matplotlib.offsetbox import AnchoredText
from datetime import datetime
import pdb
import matplotlib.pyplot as plt
import pickle
import numpy as np
import logging

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class region(object):
    """docstring for ."""

    def __init__(self, **kwargs):

        allowed_keys = set(['label', 'area', 'axis_major_length', 'axis_minor_length', \
                        'bbox', 'bbox_area', 'centroid', 'local_centroid', 'coords', 'extent', \
                        'eccentricity', 'max_intensity', 'min_intensity', \
                        'norm_max_intensity', 'norm_min_intensity', 'x_max', 'y_max', \
                        'level', 'x_min', 'y_min', 'moments', 'moments_central', \
                        'moments_hu', 'moments_normalized', 'moments_weighted', 'moments_weighted_central', \
                        'moments_weighted_hu', 'moments_weighted_normalized', 'orientation', 'perimeter' ])
        # initialize all allowed keys to false
        self.__dict__.update((key, False) for key in allowed_keys)
        # and update the given keys by their given values
        self.__dict__.update((key, value) for key, value in kwargs.items() if key in allowed_keys)

    def plot(self, wavelet_datacube, label_datacube, name = None, show = True, save_path = None, **kwargs):

        fig = plt.figure()

        if name == None:
            fig.suptitle('Region n°%d' %(self.label) )
        else:
            fig.suptitle(name)

        if self.bbox == False:
            coo_region = np.where( label_datacube.array == self.label )
            min_x_axis = np.min(coo_region[0])
            max_x_axis = np.max(coo_region[0])
            min_y_axis = np.min(coo_region[1])
            max_y_axis = np.max(coo_region[1])
            self.bbox = (min_x_axis, min_y_axis, max_x_axis, max_y_axis)

        plt.imshow( wavelet_datacube.array[ self.bbox[0]:self.bbox[2], \
                                            self.bbox[1]:self.bbox[3], \
                                            np.int(self.level) ], \
                    extent = ( self.bbox[1], self.bbox[3], \
                               self.bbox[0], self.bbox[2] ), \
                    origin = 'lower', \
                    **kwargs )

        plt.plot( self.y_max, self.x_max, '+', color = 'black' )

        plt.gca().add_artist(AnchoredText( r'level = %d'%(self.level), \
                                                loc = 'upper right', \
                                                frameon=True, \
                                                pad=0.4, \
                                                borderpad=0.1, \
                                                prop=dict(size=10)))

        try:
            clip = label_datacube.array[ self.bbox[0]:self.bbox[2], \
                                           self.bbox[1]:self.bbox[3], \
                                           np.int(self.level) ]
            clip[np.where(clip != self.label)] = 0.
            plt.contour( clip, \
                    self.label, \
                    extent = ( self.bbox[1], self.bbox[3], \
                               self.bbox[0], self.bbox[2] ), \
                    levels = 0, color = 'black', linewidth = 3 )
        except:
            print('Region too small for countours.')

        if save_path != None:
            plt.savefig(save_path, format = 'pdf')

        #plt.tight_layout()
        if show == True:
            plt.show()
        else:
            plt.close()

def make_regions_full_props(wavelet_datacube, label_datacube, verbose = False):
    '''doc to do'''

    region_list = []
    conv_table = table_noise_stds(wavelet_datacube.z_size, wavelet_type = wavelet_datacube.wavelet_type)

    for level in range(0, label_datacube.z_size):

        props_skimage = regionprops(label_datacube.array[:,:,level].astype(int), wavelet_datacube.array[:,:,level])

        for region_skimage in props_skimage:

            x_max, y_max = region_skimage.coords[np.where(\
                                  wavelet_datacube.array[region_skimage.coords[:,0], \
                                                         region_skimage.coords[:,1], \
                                                         level] == region_skimage.max_intensity)][0]


            x_min, y_min = region_skimage.coords[np.where(\
                                  wavelet_datacube.array[region_skimage.coords[:,0], \
                                                         region_skimage.coords[:,1], \
                                                         level] == region_skimage.min_intensity)][0]


            norm_max_intensity = region_skimage.max_intensity / conv_table[level]
            norm_min_intensity = region_skimage.min_intensity / conv_table[level]

            region_list.append( region( label = region_skimage.label, \
                                        area = region_skimage.area, \
                                        bbox = region_skimage.bbox, \
                                        bbox_area = region_skimage.bbox_area, \
                                        centroid = region_skimage.centroid, \
                                        local_centroid = region_skimage.local_centroid, \
                                        coords = region_skimage.coords, \
                                        extent = region_skimage.extent, \
                                        eccentricity = region_skimage.eccentricity, \
                                        max_intensity = region_skimage.max_intensity, \
                                        min_intensity = region_skimage.min_intensity, \
                                        norm_max_intensity = norm_max_intensity, \
                                        norm_min_intensity = norm_min_intensity, \
                                        x_max = x_max, \
                                        y_max = y_max, \
                                        level = level, \
                                        x_min = x_min, \
                                        y_min = y_min ) )

    if verbose == True:
        log = logging.getLogger(__name__)
        log.info('Found %d regions of significance.' %(len(region_list)))

    return region_list

def make_regions_crucial_props(wavelet_datacube, label_datacube, n_cpus = 1, verbose = False):
    '''/!\ deprecated'''

    conv_table = table_noise_stds(wavelet_datacube.z_size, wavelet_type = wavelet_datacube.wavelet_type)

    region_list = []
    region_tab = np.zeros((np.int(np.sum(label_datacube.lab_counts)), 12), order = 'F')
    wavelet_array = np.copy(wavelet_datacube.array, order='F')
    label_array = np.copy(label_datacube.array, order='F')

    region_tab[:,3] = 1.
    region_tab[:,4] =- 1.
    region_tab[:,10] =- 1.

    local_maxima_regions_para(region_tab, \
                              wavelet_array, \
                              label_array, \
                              label_datacube.lab_counts, \
                              n_cpus, \
                              verbose )

    for i in range(region_tab.shape[0]):

        norm_max_intensity = region_tab[i, 6] / conv_table[np.int(region_tab[i, 2])]
        label = i + 1
        region_list.append( region( label = label, \
                                    area = region_tab[i, 11], \
                                    max_intensity = region_tab[i, 6], \
                                    norm_max_intensity = norm_max_intensity, \
                                    x_max = region_tab[i, 0], \
                                    y_max = region_tab[i, 1], \
                                    level = region_tab[i, 2] ) )

    return region_list

def write_regions_to_pickle(region_list, filename, overwrite = True):

    if overwrite == True:
        mode = 'wb'
    else:
        mode = 'ab'

    with open(filename, mode) as outfile:
        pickle.dump(region_list, outfile)

def read_regions_from_pickle(filename):

    with open(filename, 'rb') as infile:
        region_list = pickle.load(infile)

    return region_list

if __name__ == '__main__':

    from astropy.io import fits
    from datacube import *

    acdc = coarse_datacube.from_fits('/home/ellien/devd/tests/A1365.rebin.acdc.fits')
    cdc = coarse_datacube.from_fits('/home/ellien/devd/tests/A1365.rebin.cdc.fits')
    awdc = wavelet_datacube.from_fits('/home/ellien/devd/tests/A1365.rebin.awdc.fits')
    wdc = wavelet_datacube.from_fits('/home/ellien/devd/tests/A1365.rebin.wdc.fits')
    sdc = support_datacube.from_fits('/home/ellien/devd/tests/A1365.rebin.sdc.fits')
    ldc = label_datacube.from_fits('/home/ellien/devd/tests/A1365.rebin.ldc.fits')

    rl = read_regions_from_pickle('/home/ellien/devd/tests/A1365.rebin.reglist.pkl')
    rl.sort(key = lambda x: x.norm_max_intensity, reverse = True)
    for r in rl[:10]:r.plot(wdc, ldc, cmap = 'hot')
    #region_list_f = make_regions_crucial_props(wdc, ldc, n_cpus = 4, verbose = True)
    #region_list[10001].plot(wdc, ldc, cmap = 'hot')
    #region_list_f[10001].plot(wdc, ldc, cmap = 'hot')
    #write_regions_to_pickle(region_list, '/home/ellien/devd/tests/A1365.rebin.reglist.pkl', overwrite = True)
    #region_list_test = read_regions_from_pickle('/home/ellien/devd/tests/A1365.rebin.reglist.pkl')



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# this file is part of dawis 3.0.
#
# name : label_regions.py
# author : amael ellien
# last modification : 05/2019
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# labels every regions of the support. it means that a number is given to each
# region of the support (packs of pixels with a value of 1 surrounded by pixels
# with a value of 0), starting at 1. note that the labbeling is made so each
# region is unique through the whole wavelet space.
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

from dawis.datacube import *
from skimage.measure import label
import numpy as np
import matplotlib.pyplot as plt
import logging

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def label_regions(support_datacube, verbose = False):

    '''
    labelling of the regions (pixels surrounded by non-zero pixels) of the
    multiscale support. note that the labbeling is made so each region is unique
    through the whole wavelet space.
    '''

    label_array = np.zeros(support_datacube.shape, order = 'f')
    lab_counts = np.zeros(support_datacube.z_size)

    if verbose == True:
        log = logging.getLogger(__name__)
        log.info('\nLabelling regions')

    for level in range(0, support_datacube.z_size):
        label_array[:, :, level], lab_counts[level] = label(support_datacube.array[:,:,level], \
                                                      connectivity = 1, \
                                                      return_num = True)
        if level != 0:
            reg = np.where( label_array[:, :, level] != 0 )               # here we make
            temp = label_array[:, :, level]                           # sure that the
            temp[reg] = temp[reg] + np.sum( lab_counts[0:level] ) # labels are
            label_array[:, :, level] = temp                           # unique through
                                                            # the whole
                                                            # wavelet space.
        if verbose == True:
            log.info('level %02d nregions %d labels %d - %d'\
                    %(level+1,lab_counts[level],\
                    np.max(label_array[:,:,level])-lab_counts[level]+1,\
                    np.max(label_array[:,:,level])))

    del temp, reg

    label_dc = label_datacube(label_array, lab_counts, support_datacube.fheader)

    return label_dc

if __name__ == '__main__':

    from atrous import atrous
    from numpy.random import normal
    from astropy.io import fits
    from wavelet_transforms import bspl_atrous, wavelet_datacube
    from hard_threshold import hard_threshold

    hdu = fits.open('/home/ellien/wavelets/A1365.rebin.fits')
    im = hdu[0].data
    header = hdu[0].header
    cdc, wdc = bspl_atrous(im, 10, header)
    sdc = hard_threshold(wdc, n_sigmas = 5)
    ldc = label_regions(sdc)
    ldc.label_counts_plot()



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# This file is part of DAWIS (Detection Algorithm for Intracluster light Studies).
# Author: Amaël Ellien
# Last modification: 07/2021
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# MODULES

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import SymLogNorm
from astropy.io import fits
from dawis.exceptions import DawisDimensionError
import pdb
import copy
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class datacube(object):
    """Base class for DAWIS datacubes. Datacubes can either be wavelet planes,
    label datacubes or multiresolution supports. Each datacube comes in with met-
    hods allowing an easy way of displaying the data and save/read it into fits
    files.

    A datacube instance can either be created by declaring a Numpy array and as-
    signing a datacube type to it (a fits header created ahead of the declararion
    can also be given as an argument), or by reading directly a fits file.

    This class is created mainly for error-control purposes.

    Arguments
    ----------------

            array       The actual data, which is a Numpy array with three dim-
                        ensions. The dtype of the array, as well as the order in
                        which the elements of the array are stored (C or Fortran)
                        is left to the choice of the user. Bad dimensions will
                        raise an error. # Numpy array

            dctype      Datacube type of the datacube. The four different datacu-
                        be types are 'NONE', 'COARSE', 'WAVELET', 'LABELS' and 'SUP-
                        PORT'. # String

            fheader     Fits header associated to the datacube. The attributes
                        of the datacube are written in this fits header. If the
                        fits header was given as an argument, the attributes are
                        simply added at the end of the header, and do not overw-
                        rite any previous information. This header is then used
                        by the method 'to_fits()' when writting the datacube to
                        a fits file. # Header Object Instance

    Examples:

        >>> dc1 = dawis.datacube( np.ones(100, 100, 10), dctype = 'NONE' )
        >>> from astropy.io import fits
        >>> h2  = fits.Header()
        >>> h2['TEST_ARG'] = 10
        >>> dc2 = dawis.datacube( np.ones(100, 100, 10), dctype = 'NONE', header = h2 )

    Attributes
    ----------------

            array       See arguments.

            dctype      See arguments.

            fheader     See arguments.

            shape       Shape of the Numpy array associated to the datacube.

            x_size      Size of the first dimension (X axis, Python order).

            y_size      Size of the second dimension (Y axis, Python order).

            z_size      Size of the third dimension (Z axis, Python order).

    Useful Methods
    ----------------

            to_fits()   Write the datacube and its header to a fits file.

            from_fits() Create a datacube instance and the associated header from
                        a fits file.

            show()      Display information about the datacube.

            help()      Print this help.

    """


    def __init__(self, array, dctype = 'NONE', fheader = None):

        # ATTRIBUTES
        self.array  = array
        ndim = np.ndim(self.array)
        if ndim != 3:
            raise DawisDimensionError('Not a datacube -->', ndim, 3)

        self.shape  = np.shape(array)
        self.dctype = dctype
        self.x_size, self.y_size, self.z_size = self.shape

        # FITS HEADER
        if fheader == None:
            self.fheader = fits.Header()
        else:
            self.fheader = copy.copy(fheader)
        self.fheader['DCTYPE'] = self.dctype
        self.fheader.comments['DCTYPE'] = 'DAWIS datacube type'
        self.fheader['X_SIZE'] = self.x_size
        self.fheader.comments['X_SIZE'] = 'DAWIS X axis size (python order)'
        self.fheader['Y_SIZE'] = self.y_size
        self.fheader.comments['Y_SIZE'] = 'DAWIS Y axis size (python order)'
        self.fheader['Z_SIZE'] = self.z_size
        self.fheader.comments['Z_SIZE'] = 'DAWIS Z axis size (python order)'


    def to_fits(self, filename, overwrite = False):
        '''Write the datacube to a fits file. The fits header associated to the
        datacube is automatically used as header for the new fits file. Note that
        this function only set a Primary Header Data Unit. The output fits file
        will therefore always feature one block only.

        Arguments
        ----------------

                filename        Path/name of the fits file in which the datacube
                                is written. # String

                overwrite       Optional argument (default is False). If value
                                set to True, DAWIS will overwrite the file if it
                                already exists. # Boolean

        Examples:

            >>> dc1 = dawis.datacube( np.ones(100, 100, 10), dctype = 'NONE' )
            >>> dc1.to_fits('dc1.fits')
        '''

        if self.fheader['DCTYPE'] == 'LABEL':
            hdu_dc = fits.PrimaryHDU(self.array, header = self.fheader)
            hdu_lab = fits.ImageHDU(self.lab_counts)
            hdu = fits.HDUList([hdu_dc, hdu_lab])
            hdu.writeto(filename, overwrite = overwrite)
        else:
            hdu = fits.PrimaryHDU(self.array, header = self.fheader)
            hdu.writeto(filename, overwrite = overwrite)


    @classmethod
    def from_fits(cls, filename):
        '''Create a datacube instance from a fits file. Note that only the Prima-
        ry Header Data Unit is read. The associated header is then used as header
        attribute for the datacube instance. If there are more blocks in the fits
        file, the other are ignored.

        Arguments
        ----------------

                filename        Path/name of the fits file which is read. # String

        Examples:

            >>> dc2.from_fits('dc1.fits') # Here 'dc1.fits' is a fits file
                                          # containing a datacube.
        '''

        hdu = fits.open(filename)
        fheader, data = hdu[0].header, hdu[0].data
        try:
            dctype = fheader['DCTYPE']
            if dctype in ['WAVELET', 'COARSE']:
                return cls(data, fheader['WAVTYPE'], fheader)
            elif dctype == 'SUPPORT':
                return cls(data, fheader['TRSHTYPE'], fheader)
            elif dctype == 'LABEL':
                lab_counts = hdu[1].data
                return cls(data, lab_counts, fheader)
        except:
            dctype = 'NONE'

        return cls(data, dctype, fheader)


    def show(self):
        '''Display useful information about the datacube. Version for base data-
        cube is fairly useless.

        Examples:

            >>> dc = dawis.datacube( np.zeros((100,100,10), dctype = 'NONE') )
            >>> dc.show()
        '''
        print('\ndctype : %s\nshape : %d x %d x %d\n' %(self.dctype, \
                                                        self.x_size, \
                                                        self.y_size, \
                                                        self.z_size) )

    def waveplot(self, ncol = 5, name = None, show = True, save_path = None, **kwargs ):
        '''Display the datacube under the form a grid of plots. This function
        uses matplotlib.pyplot.imshow() to plot each plane of which the key args
        can be passed as args here.

        Arguments
        ----------------

                ncol        Number of columns in the grid of plots. # Int

                name        Title of the matplotlib figure. # String or None

                **kwargs    The matplotlib kwargs. See matplotlib.pyplot.imshow()
                            documentation for more information.
        Examples:

            >>> from numpy.random import normal
            >>> dc3 = dawis.datacube( normal(0, 1, (100, 100, 10)), dctype = 'NONE' )
            >>> dc3.waveplot(name = 'Datacube of random values')
            >>> dc3.waveplot(ncol = 2)
            >>> dc3.waveplot(name = 'Datacube of random values with bicubic\
                        interpolation', cmap = 'hot', interpolation = 'bicubic')
        '''

        # Plot make-up.
        if self.z_size > ncol:
            nline = np.int( np.ceil(self.z_size / ncol) )
        else:
            nline = 1

        fig, grid = plt.subplots(nline, ncol, sharex = False, sharey = False, \
                                  gridspec_kw = {'hspace': 0.2, 'wspace': 0.2})

        if name == None:
            fig.suptitle('Datacube of type %s'%(self.dctype) )
        else:
            fig.suptitle(name)

        lvl = 0
        for i in range(nline):
            for j in range(ncol):
                lvl += 1
                if lvl > self.z_size:
                    if nline == 1:
                        grid[j].set_frame_on(False)
                        grid[j].get_xaxis().set_ticks([])
                        grid[j].get_yaxis().set_ticks([])
                    else:
                        grid[i][j].set_frame_on(False)
                        grid[i][j].get_xaxis().set_ticks([])
                        grid[i][j].get_yaxis().set_ticks([])
                else:
                    if nline == 1:
                        ax = grid[j]
                    else:
                        ax = grid[i][j]
                    if lvl == 1:
                        norm = ax.imshow(self.array[:, :, lvl - 1], **kwargs)
                    else:
                        im = ax.imshow(self.array[:, :, lvl - 1], **kwargs)
                    cax = plt.colorbar(norm, ax = ax, aspect = 10, pad = 0, \
                                                    orientation = 'horizontal')
                    ax.set_title('z = %d' %(lvl))
                    ax.xaxis.tick_top()
                    if j != 0:
                        ax.get_yaxis().set_ticks([])
                    if i != 0:
                        ax.get_xaxis().set_ticks([])

        if save_path != None:
            plt.savefig(save_path, format = 'png')

        plt.tight_layout()
        if show == True:
            plt.show()
        else:
            plt.close()

    def help(self):
        print(help(datacube))

    def __str__(self):
        return('\nDAWIS datacube object. Use "show()" method for more details.')

class wavelet_datacube(datacube):
    """docstring for ."""

    def __init__(self, array, wavelet_type, fheader = None):

        # Inheritance from parent class datacube
        super().__init__(array, 'WAVELET', fheader)
        self.fheader['WAVTYPE'] = wavelet_type
        self.fheader.comments['WAVTYPE'] = 'DAWIS wavelet type'
        self.wavelet_type = wavelet_type

    def std_plot(self, name = None, show = True, save_path = None, **kwargs ):
        '''doc to do'''

        fig = plt.figure()

        if name == None:
            fig.suptitle('Standard deviation against scale')
        else:
            fig.suptitle(name)

        plt.plot(np.std(self.array, axis = (0,1)), **kwargs)
        plt.yscale('log', basey = 10)

        if save_path != None:
            plt.savefig(save_path, format = 'pdf')

        #plt.tight_layout()
        if show == True:
            plt.show()
        else:
            plt.close()

    def waveplot(self, ncol = 5, name = None, show = True, save_path = None, **kwargs ):
        '''Display the datacube under the form a grid of plots. This function
        uses matplotlib.pyplot.imshow() to plot each plane of which the key args
        can be passed as args here. Version specific to datacube of type 'WAVELET'.

        Arguments
        ----------------

                ncol        Number of columns in the grid of plots. # Int

                name        Title of the matplotlib figure. # String or None

                **kwargs    The matplotlib kwargs. See matplotlib.pyplot.imshow()
                            documentation for more information.
        Examples:

            >>> from numpy.random import normal
            >>> dc3 = dawis.datacube( normal(0, 1, (100, 100, 10)), dctype = 'NONE' )
            >>> dc3.waveplot(name = 'Datacube of random values')
            >>> dc3.waveplot(ncol = 2)
            >>> dc3.waveplot(name = 'Datacube of random values with bicubic\
                        interpolation', cmap = 'hot', interpolation = 'bicubic')
        '''

        # Plot make-up.
        if self.z_size > ncol:
            nline = np.int( np.ceil(self.z_size / ncol) )
        else:
            nline = 1
            ncol = self.z_size

        fig, grid = plt.subplots(nline, ncol, sharex = False, sharey = False, \
                                  gridspec_kw = {'hspace': 0.2, 'wspace': 0.2})

        if name == None:
            fig.suptitle('Datacube of type %s'%(self.dctype) )
        else:
            fig.suptitle(name)

        lvl = 0
        for i in range(nline):
            for j in range(ncol):
                lvl += 1
                if lvl > self.z_size:
                    if nline == 1:
                        grid[j].set_frame_on(False)
                        grid[j].get_xaxis().set_ticks([])
                        grid[j].get_yaxis().set_ticks([])
                    else:
                        grid[i][j].set_frame_on(False)
                        grid[i][j].get_xaxis().set_ticks([])
                        grid[i][j].get_yaxis().set_ticks([])
                else:
                    if nline == 1:
                        ax = grid[j]
                    else:
                        ax = grid[i][j]

                    try:
                        norm = kwargs.pop('norm')
                        if lvl == 1:
                            im_norm = ax.imshow(self.array[:, :, lvl - 1], **kwargs)
                        else:
                            im = ax.imshow(self.array[:, :, lvl - 1], **kwargs)
                        cax = plt.colorbar(im_norm, ax = ax, aspect = 10, pad = 0, \
                                                        orientation = 'horizontal')
                    except:
                        logthresh = 1E-3
                        norm = SymLogNorm(logthresh)
                        if lvl == 1:
                            im_norm = ax.imshow(self.array[:, :, lvl - 1], norm = norm, **kwargs)
                            maxlog = np.log10( np.ceil( np.max( self.array[:, :, lvl - 1] )))
                            minlog = np.log10( np.ceil( - np.min( self.array[:, :, lvl - 1] )))
                        else:
                            im = ax.imshow(self.array[:, :, lvl - 1], norm = norm, **kwargs)

                        tick_locations = (list(- np.logspace(1, minlog, num = 2, base = 10)) \
                                            + [0] \
                                            + list(np.logspace(1, maxlog, num = 2, base = 10)))
                        cax = plt.colorbar(im_norm, ax = ax, aspect = 10, pad = 0, \
                                                    orientation = 'horizontal', \
                                                    ticks = tick_locations)
                    ax.set_title('z = %d' %(lvl))
                    ax.xaxis.tick_top()
                    if j != 0:
                        ax.get_yaxis().set_ticks([])
                    if i != 0:
                        ax.get_xaxis().set_ticks([])

        if save_path != None:
            plt.savefig(save_path, format = 'pdf')

        #plt.tight_layout()
        if show == True:
            plt.show()
        else:
            plt.close()

class coarse_datacube(datacube):
    """docstring for ."""

    def __init__(self, array, wavelet_type, fheader = None):

        # Inheritance from parent class datacube
        super().__init__(array, 'COARSE', fheader)
        # Attributes
        self.fheader['WAVTYPE'] = wavelet_type
        self.fheader.comments['WAVTYPE'] = 'DAWIS wavelet type'
        self.wavelet_type = wavelet_type

class support_datacube(datacube):
    """docstring for ."""

    def __init__(self, array, threshold_type, noise_pixels, fheader = None):

        # Inheritance from parent class datacube
        super().__init__(array, 'SUPPORT', fheader)
        # Attributes
        self.threshold_type = threshold_type
        self.noise_pixels = noise_pixels
        self.fheader['TRSHTYPE'] = threshold_type
        self.fheader.comments['TRSHTYPE'] = 'DAWIS threshold type'

    def histogram_noise(self, name = None, show = True, save_path = None, **kwargs):
        '''doc to do'''

        fig = plt.figure()

        if name == None:
            fig.suptitle('Datacube of type %s'%(self.dctype) )
        else:
            fig.suptitle(name)

        plt.hist(self.noise_pixels.flatten(),  bins = 'auto',  \
                                               histtype = 'bar', \
                                               rwidth = 10. , \
                                               align='left', \
                                               **kwargs)

        if save_path != None:
            plt.savefig(save_path, format = 'png')

        #plt.tight_layout()
        if show == True:
            plt.show()
        else:
            plt.close()

    def test_gaussianity(self, alpha = 1E-3):
        k, pval = normaltest( self.noise_pixels )
        if pval > alpha:
            print('/!\ Warning /!\ noise does not look Gaussian : p = %1.5f alpha = %1.5f'%(pval, alpha))
        else:
            print('Gaussianity test passed : p = %1.5f alpha = %1.5f'%(pval, alpha))
        return k, pval

class label_datacube(datacube):
    """docstring for ."""

    def __init__(self, array, lab_counts, fheader = None):

        # Inheritance from parent class datacube
        super().__init__(array, 'LABEL', fheader)
        # Attributes
        self.lab_counts = lab_counts
        self.fheader['LABC'] = np.sum(self.lab_counts)
        self.fheader.comments['LABC'] = 'DAWIS total label counts'

    def label_counts_plot(self, name = None, show = True, save_path = None, **kwargs ):
        '''doc to do'''

        fig = plt.figure()

        if name == None:
            fig.suptitle('Label counts against scale')
        else:
            fig.suptitle(name)

        plt.plot(self.lab_counts, **kwargs)

        if save_path != None:
            plt.savefig(save_path, format = 'pdf')

        #plt.tight_layout()
        if show == True:
            plt.show()
        else:
            plt.close()

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

if __name__ == '__main__':

    from numpy.random import normal
    dc3 = datacube( normal(0, 1,(100, 100, 9)), dctype = 'NONE' )
    dc3.waveplot(name = 'Datacube of random values')
    #dc3.waveplot(ncol = 2)
    #dc3.waveplot(cmap = 'hot', interpolation = 'bicubic')
    #FILE = '/home/ellien/devd/tests/datacube_test.fits'

    #dc_test.to_fits(FILE, overwrite = True )

    #dc_test_3 = datacube( np.ones( (10, 10, 3) ), dctype = 'WAVELET DATACUBE' )
    #print(dc_test_3.fheader['DCTYPE'])
    #print(dc_test_3.fheader)

    #dc_test_4 = datacube( np.ones( (10, 10, 3) ), dctype = 'BULLSHIT' )

    #dc_test_2 = datacube.from_fits(FILE)
    #print(dc_test_2.fheader)



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# This file is part of DAWIS.
#
# NAME : restore_objects.py
# AUTHOR : Amael Ellien
# LAST MODIFICATION : 07/2021
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

from datetime import datetime
import pdb
import matplotlib.pyplot as plt
import pickle
import numpy as np
import ray
import logging

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class restored_object(object):
    """docstring for restored_object."""

    def __init__(self, image, bbox, level, eccentricity, filter_kw, flag_convergence, sum_wr, norm_wr):
        self.bbox = bbox
        self.image = image
        self.level = level
        self.eccentricity = eccentricity
        self.filter = filter_kw
        self.flag_convergence = flag_convergence
        self.sum_wr = sum_wr
        self.norm_wr = norm_wr

def write_objects_to_pickle(object_list, filename, overwrite = True):

    if overwrite == True:
        mode = 'wb'
    else:
        mode = 'ab'

    with open(filename, mode) as outfile:
        pickle.dump(object_list, outfile)

def read_objects_from_pickle(filename):

    with open(filename, 'rb') as infile:
        object_list = pickle.load(infile)

    return object_list


@ray.remote
def restore_patch(interscale_tree_patch, wavelet_datacube, label_datacube, extent_sep, ecc_sep, lvl_sep_lin, lvl_sep_big):

    object_patch = []
    for tree in interscale_tree_patch:

        image, filter_kw, flag_convergence, sum_wr, norm_wr = restore_object( tree, \
                                                                              wavelet_datacube, \
                                                                              label_datacube, \
                                                                              extent_sep, \
                                                                              ecc_sep, \
                                                                              lvl_sep_lin, lvl_sep_big )

        object_patch.append(restored_object(image, tree.bbox, \
                                            tree.interscale_maximum.level, \
                                            tree.interscale_maximum.eccentricity, \
                                            filter_kw, \
                                            flag_convergence, \
                                            sum_wr, \
                                            norm_wr ))

    return object_patch

def restore_object( interscale_tree, wavelet_datacube, label_datacube, extent_sep, ecc_sep, lvl_sep_lin, lvl_sep_big, verbose = False ):

    bspl = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
    haar = 1 / 2. * np.array([ 1, 0, 1 ])

    flag_convergence = True

    if (( interscale_tree.extent < extent_sep ) or (  interscale_tree.eccentricity >= ecc_sep )) & ( interscale_tree.interscale_maximum.level < lvl_sep_lin ) :
        image, sum_wr, norm_wr = interscale_tree.CG_minimization( wavelet_datacube, label_datacube, filter = haar, \
                                                        synthesis_operator = 'ADJOINT', verbose = verbose )
        filter_kw = 'HAAR'

    else:
        image, sum_wr, norm_wr = interscale_tree.CG_minimization( wavelet_datacube, label_datacube, filter = bspl, \
                                                    synthesis_operator = 'ADJOINT', verbose = verbose )
        filter_kw = 'BSPL'

    # Security
    if np.isnan(np.sum(image)):
        image = np.zeros( image.shape )
        flag_convergence = False

    return image, filter_kw, flag_convergence, sum_wr, norm_wr


def restore_objects_default(interscale_tree_list, wavelet_datacube, label_datacube, lvl_sep_big, extent_sep, ecc_sep, lvl_sep_lin, size_patch_small = 50, size_patch_big = 5, size_big_objects = 512, n_cpus = 1 ):

    if (len(interscale_tree_list) < size_patch_big) or (n_cpus == 1):
        # params_minimization = strategy(tree)
        object_list = []
        bspl = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
        logging.info('Size tree patch (%d) lower than %d, not parallelizing.'%(len(interscale_tree_list),size_patch_big ))
        for tree in interscale_tree_list:
            image, filter_kw, flag_convergence, sum_wr, norm_wr = restore_object( tree, \
                                                                                  wavelet_datacube, \
                                                                                  label_datacube, \
                                                                                  extent_sep, \
                                                                                  ecc_sep, \
                                                                                  lvl_sep_lin, \
                                                                                  lvl_sep_big, \
                                                                                  verbose = True )

            object_list.append( restored_object(image, tree.bbox, \
                                                tree.interscale_maximum.level, \
                                                tree.interscale_maximum.eccentricity, \
                                                filter_kw, \
                                                flag_convergence, \
                                                sum_wr, \
                                                norm_wr ))

    else:
        logging.info('Size tree patch (%d) greater than %d, activating Ray store.'%(len(interscale_tree_list),size_patch_big ))
        ray.init(num_cpus = n_cpus)
        id_wdc = ray.put(wavelet_datacube)
        id_ldc = ray.put(label_datacube)
        id_extent_sep = ray.put(extent_sep)
        id_ecc_sep = ray.put(ecc_sep)
        id_lvl_sep_lin = ray.put(lvl_sep_lin)
        id_lvl_sep_big = ray.put(lvl_sep_big)

        interscale_tree_list.sort(key = lambda x: x.interscale_maximum.area, reverse = True)

        small_interscale_tree_patch = []
        small_object_patch = []
        big_interscale_tree_patch = []
        big_object_patch = []

        for tree in interscale_tree_list:

            if tree.x_size > size_big_objects & tree.y_size > size_big_objects:
                big_interscale_tree_patch.append(tree)

                if len(big_interscale_tree_patch) >= size_patch_big:
                    big_object_patch.append( restore_patch.remote( big_interscale_tree_patch, id_wdc, id_ldc, id_extent_sep, id_ecc_sep, id_lvl_sep_lin, id_lvl_sep_big))
                    big_interscale_tree_patch = []

            else:

                if big_interscale_tree_patch:
                        big_object_patch.append( restore_patch.remote( big_interscale_tree_patch, id_wdc, id_ldc, id_extent_sep, id_ecc_sep, id_lvl_sep_lin, id_lvl_sep_big ))

                small_interscale_tree_patch.append(tree)

                if len(small_interscale_tree_patch) >= size_patch_small:
                    small_object_patch.append( restore_patch.remote( small_interscale_tree_patch, id_wdc, id_ldc, id_extent_sep, id_ecc_sep, id_lvl_sep_lin, id_lvl_sep_big ))
                    small_interscale_tree_patch = []

        if small_interscale_tree_patch:
            small_object_patch.append( restore_patch.remote( small_interscale_tree_patch, id_wdc, id_ldc, id_extent_sep, id_ecc_sep, id_lvl_sep_lin, id_lvl_sep_big ))

        object_list = []
        for id_patch in big_object_patch:
            patch = ray.get( id_patch )
            object_list = object_list + patch
        for id_patch in small_object_patch:
            patch = ray.get( id_patch )
            object_list = object_list + patch

        ray.shutdown()

    return object_list

if __name__ == '__main__':

    from make_regions import *
    from datacube import *
    from matplotlib.colors import SymLogNorm, LogNorm
    from astropy.io import fits
    from atrous import atrous
    from make_regions import make_regions_full_props
    from wavelet_transforms import bspl_atrous, wavelet_datacube
    from hard_threshold import hard_threshold
    from label_regions import label_regions
    from anscombe_transforms import *
    from skimage.metrics import structural_similarity as ssim
    from make_interscale_trees import make_interscale_trees

    hdu = fits.open('/home/ellien/devd/gallery/A1365.rebin.fits')
    im = hdu[0].data
    header = hdu[0].header
    n_levels = np.int(np.min(np.floor(np.log2(im.shape))))
    print(im.shape, n_levels, 2**n_levels)
    n_levels = 3

    sigma, mean, gain = pg_noise_bissection(im, max_err = 1E-3, n_sigmas = 3)
    aim = anscombe_transform(im, sigma, mean, gain)
    acdc, awdc = bspl_atrous(aim, n_levels, header)
    sdc = hard_threshold(awdc, n_sigmas = 5)
    ldc = label_regions(sdc)
    cdc, wdc = bspl_atrous(im, n_levels, header)
    rl = make_regions_full_props(wdc, ldc, verbose = True)
    itl = make_interscale_trees(rl, ldc, tau = 0.8, min_span = 3, verbose = True)
    ol = restore_objects_default(itl, wdc,ldc, size_patch_small = 50, \
                                               size_patch_big = 5, \
                                               size_big_objects = 512, \
                                               n_cpus = 4 )



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# This file is part of DAWIS.
#
# NAME : gif.py
# AUTHOR : Amael Ellien
# LAST MODIFICATION : 07/2021
# ffmpeg -framerate 5 -i img-%02d.png video.avi

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

from matplotlib.colors import LogNorm
from subprocess import run
import matplotlib.pyplot as plt
import logging
logging.getLogger('matplotlib').setLevel(logging.WARNING)
import numpy as np
from astropy.visualization import LinearStretch, LogStretch
from astropy.visualization import ZScaleInterval, MinMaxInterval
from astropy.visualization import ImageNormalize
from dawis.restore_objects import *
from astropy.io import fits

def plot_frame( level, it, nobj, original_image, restored_image, residuals, atom, outpath = '' ):

    '''doc to do'''

    fig = plt.figure(dpi = 500)
    ax = fig.subplots(2, 2, sharex = True, sharey = True, \
                                gridspec_kw = { 'hspace': 0.1, 'wspace': 0.1 })
    fig.suptitle('Level %d\nIteration %d\n%d objects' %(level, it, nobj))

    axim = ax[0][0].imshow(original_image,  norm = ImageNormalize( original_image, \
                                            interval = ZScaleInterval(), \
                                            stretch = LinearStretch()), \
                                            origin = 'lower', \
                                            cmap = 'binary' )
    cax = plt.colorbar(axim, ax = ax[0][0], aspect = 10, pad = 0, \
                                    orientation = 'vertical')
    cax.ax.tick_params(labelsize = 5)
    ax[0][0].set_title('Original')

    axim = ax[0][1].imshow(residuals, norm = ImageNormalize( original_image, \
                                      interval = ZScaleInterval(),
                                      stretch = LinearStretch()), \
                                      origin = 'lower', \
                                      cmap = 'binary' )

    cax = plt.colorbar(axim, ax = ax[0][1], aspect = 10, pad = 0, \
                                    orientation = 'vertical')
    cax.ax.tick_params(labelsize = 5)
    ax[0][1].set_title('Residuals')

    axim = ax[1][0].imshow(restored_image,  norm = ImageNormalize( original_image, \
                                            interval = ZScaleInterval(),
                                            stretch = LinearStretch()), \
                                            origin = 'lower', \
                                            cmap = 'binary' )

    cax = plt.colorbar(axim, ax = ax[1][0], aspect = 10, pad = 0, \
                                    orientation = 'vertical')
    cax.ax.tick_params(labelsize = 5)
    ax[1][0].set_title('Restored')

    axim = ax[1][1].imshow(atom, norm = ImageNormalize( atom, \
                                 interval = MinMaxInterval(),
                                 stretch = LogStretch()), \
                                 origin = 'lower', \
                                  cmap = 'binary' )

    cax = plt.colorbar(axim, ax = ax[1][1], aspect = 10, pad = 0, \
                                    orientation = 'vertical')
    cax.ax.tick_params(labelsize = 5)
    ax[1][1].set_title('Atom')

    plt.savefig(''.join(( outpath, '.frame.it%03d.png' %(it) )), format = 'png' )
    plt.close()

def make_gif( framerate, outpath ):

    '''doc to do'''

    try:
        args_ffmpeg = [ 'ffmpeg', '-framerate', str(framerate), '-y', '-i', \
                ''.join(( outpath, '.frame.it%03d.png' )), \
                ''.join(( outpath, '.run.avi' )) ]

        run( args_ffmpeg )
    except:
        logging.info('Could not create frame gif')

    #try:
    #    args_ffmpeg = [ 'ffmpeg', '-framerate', str(framerate), '-i', \
    #            ''.join(( outpath, '.hist.it%03d.png' )), \
    #            ''.join(( outpath, '.hist.avi' )) ]

    #    run( args_ffmpeg )

    #except:
    #    logging.info('Could not create noise hist gif')

    #args_rm = [ 'rm', ''.join(( outpath, '.frame*.png' )) ]
    #run( args_rm )

def remake_frames( imagepath, outpath, gamma, n_it ):

    #outpath = '/home/ellien/devd/tests/full/tidal_group_f814w_sci.rl.it'


    image = fits.getdata(imagepath)
    res = np.copy(image)
    rec = np.zeros(image.shape)
    for it in range(1, n_it + 1):
        atom = np.zeros(image.shape)
        ol = read_objects_from_pickle(outpath+'.frame.it%03d'%(it)+'.pkl')
        for object in ol:
            x_min, y_min, x_max, y_max = object.bbox
            atom[ x_min : x_max, y_min : y_max ] += object.image * gamma
        res -= atom
        rec += atom
        level = np.max( [ x.level for x in ol ] )
        plot_frame( level = level, it = it, nobj = len(ol), original_image = image,  \
                                    restored_image = rec, residuals = res, \
                                    atom = atom, outpath = outpath )

def plot_convergence( cparl, outpath, **kwargs ):
    '''doc to do'''

    level = []
    it = []
    avnormeps = []
    normeps = []
    eps = []
    flux_rec = []
    flux_res = []
    nobj = []
    time = []

    for cpar in cparl:

        level.append(int(cpar[0]))
        it.append(int(cpar[1]))
        avnormeps.append(float(cpar[2]))
        normeps.append(float(cpar[3]))
        eps.append(float(cpar[4]))
        flux_rec.append(float(cpar[5]))
        flux_res.append(float(cpar[6]))
        nobj.append(int(cpar[7]))
        time.append(cpar[8])

    fig = plt.figure(figsize = (9, 6), dpi = 500)
    plt.plot(it, avnormeps, label = 'window normalized eps', **kwargs)
    plt.plot(it, normeps, label = 'normalized eps', **kwargs)
    plt.plot(it, eps, label = 'eps', **kwargs)
    plt.plot(it, flux_rec, label = 'flux restored fraction', **kwargs)
    plt.plot(it, flux_res, label = 'flux residuals fraction', **kwargs)
    plt.grid(linestyle='-', linewidth=2, alpha=0.2)

    plt.legend()
    plt.yscale('log')
    plt.savefig(''.join(( outpath, '.conv.pdf' %(it) )), format = 'pdf' )
    plt.close()



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# This file is part of DAWIS (Detection Algorithm for Intracluster light Studies).
# Author: Amaël Ellien
# Last modification: 20/05/2021
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# MODULES
import numpy as np
from dawis.datacube import datacube
from dawis.exceptions import *
import pdb
from datetime import datetime
from scipy.ndimage import gaussian_filter
from scipy.signal import convolve2d
from scipy.signal import bspline
import matplotlib.pyplot as plt
from dawis.congrid import congrid
import logging

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

def boundary_conditions(position_array, sizes, conditions = 'prolongation' ):
    '''
    Border conditions, doc to do
    '''

    if np.ndim(position_array) not in [1, 3]:
        raise DawisDimensionError('Wrong dimensions for array -->', np.ndim(position_array), [1, 3])

    if np.ndim(sizes) not in [0, 1]:
        raise DawisDimensionError('Wrong dimensions for array -->', np.ndim(sizes), [0, 1])

    if type(conditions) != str:
        raise DawisWrongType('Wrong type -->', type(conditions), 'str')

    allowed_conditions = ['mirror', 'prolongation', 'loop' ]

    if conditions not in allowed_conditions:
        raise DawisValueError('Not an allowed condition -->', condition, allowed_conditions)

    if conditions == 'prolongation':

        position_array[np.where(position_array < 0)] = 0

        if ( np.ndim(position_array) == 3 ) & ( np.ndim(sizes) == 1 ):
            position_array[0, np.where(position_array[0,:] > sizes[0] - 1)] = sizes[0] - 1
            position_array[1, np.where(position_array[1,:] > sizes[1] - 1)] = sizes[1] - 1
        else:
            position_array[np.where(position_array > sizes - 1)] = sizes - 1

    if conditions == 'mirror':

        position_array[np.where(position_array < 0)] -= 2 * position_array[np.where(position_array < 0)]

        if ( np.ndim(position_array) == 3 ) & ( np.ndim(sizes) == 1 ):
            position_array[0][np.where(position_array[0,:] > sizes[0] - 1)] -= 2 * ( position_array[0][np.where(position_array[0,:] > sizes[0] - 1)] - sizes[0] )
            position_array[1][np.where(position_array[1,:] > sizes[1] - 1)] -= 2 * ( position_array[1][np.where(position_array[1,:] > sizes[1] - 1)] - sizes[1] )
        else:
            position_array[np.where(position_array > sizes - 1)] -= sizes - ( position_array[np.where(position_array > sizes - 1 )] - sizes )

    if conditions == 'loop':

        if ( np.ndim(position_array) == 3 ) & ( np.ndim(sizes) == 1 ):
            position_array[0][np.where(position_array[0,:] > sizes[0] - 1)] -= sizes[0]
            position_array[1][np.where(position_array[1,:] > sizes[1] - 1)] -= sizes[1]
            position_array[0][np.where(position_array[0,:] < 0)] += sizes[0]
            position_array[1][np.where(position_array[1,:] < 0)] += sizes[1]
        else:
            position_array[np.where(position_array > sizes - 1)] -= sizes
            position_array[np.where(position_array < 0)] += sizes

    return position_array

def atrous(image, n_levels, filter, conditions = 'loop', verbose = False):
    '''
    A trous algorithm, doc to do
    '''

    #===========================================================================
    # Exceptions
    if image.ndim != 2:
        raise DawisDimensionError('Not an image -->', image.ndim, 2)
    image_x_size = image.shape[0]
    image_y_size = image.shape[1]

    if filter.ndim > 2:
        raise DawisDimensionError('Not a 1D or 2D array -->', image.ndim, 2)

    if n_levels <= 0:
        raise DawisValueError('Number of wavelet planes cannot be negative -->', \
                                                n_levels, 'positive integer')

    #===========================================================================
    # Position array for filter
    if filter.ndim == 1:
        filter_ctr_index = filter.size // 2 + 1
        filter_half_size = filter.size // 2
        positive_positions = np.logspace( 0, filter_half_size - 1, \
                                            num = filter_half_size, base = 2)
        position_array = np.concatenate((- np.flip(positive_positions), np.zeros(1),
                                                            positive_positions))
    elif filter.ndim == 2:
        filter_ctr_index = (filter.shape[0] // 2 + 1, filter.shape[1] // 2 + 1 )
        filter_half_size = (filter.shape[0] // 2, filter.shape[1] // 2 )
        positive_positions_x = np.logspace( 0, filter_half_size[0] - 1, \
                                            num = filter_half_size[0], base = 2)
        positive_positions_y = np.logspace( 0, filter_half_size[1] - 1, \
                                                    num = filter_half_size[1], base = 2)
        position_array_x = np.concatenate((- np.flip(positive_positions_x), np.zeros(1),
                                                            positive_positions_x))
        position_array_y = np.concatenate((- np.flip(positive_positions_y), np.zeros(1),
                                                            positive_positions_y))
        position_array = np.array(np.meshgrid(position_array_x, position_array_y))

    #===========================================================================
    # Arrays
    coarse_array = np.zeros((image_x_size, image_y_size, n_levels + 1), dtype = np.float32)
    wavelet_array = np.zeros((image_x_size, image_y_size, n_levels ), dtype = np.float32)
    coarse_array[:,:,0] = image

    #===========================================================================
    # Convolution
    for level in range(0, n_levels):

        position_array_lvl = np.array(position_array) * np.power(2, level)
        position_array_lvl = position_array_lvl.astype(np.int)

        if filter.ndim == 2:
            with np.nditer(coarse_array[:,:,level + 1], flags = ['multi_index'], op_flags = ['readwrite']) as it:
                for coeff in it:
                    position_array_coeff = np.copy(position_array_lvl)
                    position_array_coeff[0] += it.multi_index[0]
                    position_array_coeff[1] += it.multi_index[1]
                    position_array_coeff = boundary_conditions(position_array_coeff, image.shape, conditions)

                    coeff += np.sum(coarse_array[position_array_coeff[0], position_array_coeff[1], level] * filter)

        elif filter.ndim == 1:

            temporary_copy = np.copy(coarse_array[:,:,level + 1])
            # Over X axis

            for x in range(0, image_x_size):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += x

                position_array_coeff = boundary_conditions(position_array_coeff, image.shape[0], conditions)

                temporary_copy[x,:] = np.matmul(np.array([filter]), coarse_array[position_array_coeff, :, level])

            for y in range(0, image_y_size):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += y
                position_array_coeff = boundary_conditions(position_array_coeff, image.shape[1], conditions)

                coarse_array[:,y,level + 1] = np.matmul(np.array([filter]), temporary_copy[ :, position_array_coeff].T)[0]

    for level in range(0, n_levels):
        wavelet_array[:,:,level] = coarse_array[:,:,level] - coarse_array[:,:,level+1]
    #wavelet_array[:,:,n_levels - 1] = np.copy(coarse_array[:,:,n_levels - 1])

    return coarse_array, wavelet_array

def interlaced_atrous_congrid(image, n_levels, n_voices, filter, verbose = False):

    '''
    Interlaced à trous algorithm, doc to do
    '''

    #===========================================================================
    # Exceptions
    if image.ndim != 2:
        raise DawisDimensionError('Not an image -->', image.ndim, 2)
    image_x_size = image.shape[0]
    image_y_size = image.shape[1]

    if filter.ndim > 2:
        raise DawisDimensionError('Not a 1D or 2D array -->', image.ndim, 2)

    if n_levels <= 0:
        raise DawisValueError('Number of wavelet planes cannot be negative -->', \
                                                n_levels, 'positive integer')

    #===========================================================================
    # Position array for filter
    if filter.ndim == 1:
        filter_ctr_index = filter.size // 2 + 1
        filter_half_size = filter.size // 2
        positive_positions = np.logspace( 0, filter_half_size - 1, \
                                            num = filter_half_size, base = 2)
        position_array = np.concatenate((- np.flip(positive_positions), np.zeros(1),
                                                            positive_positions))
    elif filter.ndim == 2:
        filter_ctr_index = (filter.shape[0] // 2 + 1, filter.shape[1] // 2 + 1 )
        filter_half_size = (filter.shape[0] // 2, filter.shape[1] // 2 )
        positive_positions_x = np.logspace( 0, filter_half_size[0] - 1, \
                                            num = filter_half_size[0], base = 2)
        positive_positions_y = np.logspace( 0, filter_half_size[1] - 1, \
                                                    num = filter_half_size[1], base = 2)
        position_array_x = np.concatenate((- np.flip(positive_positions_x), np.zeros(1),
                                                            positive_positions_x))
        position_array_y = np.concatenate((- np.flip(positive_positions_y), np.zeros(1),
                                                            positive_positions_y))
        position_array = np.array(np.meshgrid(position_array_x, position_array_y))

    #===========================================================================
    # Arrays
    final_coarse_array = np.zeros((image_x_size, image_y_size, 2 * n_levels  ), dtype = np.float32)
    final_wavelet_array = np.zeros((image_x_size, image_y_size, 2 * n_levels ), dtype = np.float32)

    #===========================================================================
    # Convolution 0

    anx, any = np.int(image_x_size / ( np.sqrt(2) / 2 )), \
               np.int(image_y_size / ( np.sqrt(2) / 2 ))
    print(anx, any)
    coarse_array_1 = np.zeros((image_x_size, image_y_size, n_levels + 1 ), dtype = np.float32)
    coarse_array_2 = np.zeros((image_x_size, image_y_size, n_levels ), dtype = np.float32)

    congrid_coarse_array = np.zeros((anx, \
                                     any, \
                                     n_levels + 1 ), dtype = np.float32)

    coarse_array_1[:,:,0] = np.copy(image)
    congrid_coarse_array[:,:,0] = congrid(image, (anx, any), method = 'spline')

    #===========================================================================
    # Convolution 1
    for level in range(0, n_levels ):

        position_array_lvl = np.array(position_array) * np.power(2, level)
        position_array_lvl = position_array_lvl.astype(np.int)

        if filter.ndim == 2:
            with np.nditer(coarse_array_1[:,:,level + 1], flags = ['multi_index'], op_flags = ['readwrite']) as it:
                for coeff in it:
                    position_array_coeff = np.copy(position_array_lvl)
                    position_array_coeff[0] += it.multi_index[0]
                    position_array_coeff[1] += it.multi_index[1]
                    position_array_coeff = boundary_conditions(position_array_coeff, interp_im.shape, 'loop')

                    coeff += np.sum(coarse_array_1[position_array_coeff[0], position_array_coeff[1], level] * filter)

        elif filter.ndim == 1:

            temporary_copy = np.zeros((image_x_size, image_y_size))
            # Over X axis

            for x in range(0, image_x_size ):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += x
                position_array_coeff = boundary_conditions(position_array_coeff, image.shape[0], 'loop')

                temporary_copy[x,:] = np.matmul(np.array([filter]), coarse_array_1[position_array_coeff, :, level])

            for y in range(0, image_y_size ):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += y
                position_array_coeff = boundary_conditions(position_array_coeff, image.shape[1] , 'loop')

                coarse_array_1[:,y,level + 1] = np.matmul(np.array([filter]), temporary_copy[ :, position_array_coeff].T)[0]

    # Convolution 2
    for level in range(0, n_levels ):

        position_array_lvl = np.array(position_array) * np.power(2, level)
        position_array_lvl = position_array_lvl.astype(np.int)

        if filter.ndim == 2:
            with np.nditer(coarse_array_2[:,:,level + 1], flags = ['multi_index'], op_flags = ['readwrite']) as it:
                for coeff in it:
                    position_array_coeff = np.copy(position_array_lvl)
                    position_array_coeff[0] += it.multi_index[0]
                    position_array_coeff[1] += it.multi_index[1]
                    position_array_coeff = boundary_conditions(position_array_coeff, interp_im.shape, 'loop')

                    coeff += np.sum(coarse_array_2[position_array_coeff[0], position_array_coeff[1], level] * filter)

        elif filter.ndim == 1:

            temporary_copy = np.copy(congrid_coarse_array[:,:,level + 1])

            for x in range(0, anx):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += x
                position_array_coeff = boundary_conditions(position_array_coeff, anx, 'loop')

                temporary_copy[x,:] = np.matmul(np.array([filter]), congrid_coarse_array[position_array_coeff, :, level])

            for y in range(0, any):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += y
                position_array_coeff = boundary_conditions(position_array_coeff, any, 'loop')

                congrid_coarse_array[:,y,level + 1] = np.matmul(np.array([filter]), temporary_copy[ :, position_array_coeff].T)[0]

            coarse_array_2[:,:,level] = congrid(congrid_coarse_array[:,:,level + 1], (image_x_size, image_y_size), method = 'spline')

    # Correction for mean bias arror given by interpolation
    for level in range(0, n_levels):

        interpmean = ( np.mean( coarse_array_1[:,:,level + 1] ) + np.mean( coarse_array_1[:,:,level] ) ) / 2.
        norm = interpmean - np.mean( coarse_array_2[:,:,level] )
        coarse_array_2[:,:,level] += norm

    voice = 0
    for level in range(0, n_levels):

        final_wavelet_array[:,:,voice] = coarse_array_1[:,:,level] - coarse_array_2[:,:,level]
        final_coarse_array[:,:,voice] = coarse_array_1[:,:,level]
        #print(level, voice, np.mean(final_wavelet_array[:,:,voice]), np.std(final_wavelet_array[:,:,voice]))
        voice += 1

        final_wavelet_array[:,:,voice] = coarse_array_2[:,:,level] - coarse_array_1[:,:,level+1]
        final_coarse_array[:,:,voice] = coarse_array_2[:,:,level]
        #print(level, voice, np.mean(final_wavelet_array[:,:,voice]), np.std(final_wavelet_array[:,:,voice]))
        voice += 1

    return final_coarse_array, final_wavelet_array

def interlaced_atrous_zeros(image, n_levels, n_voices, filter, verbose = False):

    '''
    Interlaced à trous algorithm, doc to do
    '''

    #===========================================================================
    # Exceptions
    if image.ndim != 2:
        raise DawisDimensionError('Not an image -->', image.ndim, 2)
    image_x_size = image.shape[0]
    image_y_size = image.shape[1]

    if filter.ndim > 2:
        raise DawisDimensionError('Not a 1D or 2D array -->', image.ndim, 2)

    if n_levels <= 0:
        raise DawisValueError('Number of wavelet planes cannot be negative -->', \
                                                n_levels, 'positive integer')

    #===========================================================================
    # Position array for filter
    if filter.ndim == 1:
        filter_ctr_index = filter.size // 2 + 1
        filter_half_size = filter.size // 2
        positive_positions = np.logspace( 0, filter_half_size - 1, \
                                            num = filter_half_size, base = 2)
        position_array = np.concatenate((- np.flip(positive_positions), np.zeros(1),
                                                            positive_positions))
    elif filter.ndim == 2:
        filter_ctr_index = (filter.shape[0] // 2 + 1, filter.shape[1] // 2 + 1 )
        filter_half_size = (filter.shape[0] // 2, filter.shape[1] // 2 )
        positive_positions_x = np.logspace( 0, filter_half_size[0] - 1, \
                                            num = filter_half_size[0], base = 2)
        positive_positions_y = np.logspace( 0, filter_half_size[1] - 1, \
                                                    num = filter_half_size[1], base = 2)
        position_array_x = np.concatenate((- np.flip(positive_positions_x), np.zeros(1),
                                                            positive_positions_x))
        position_array_y = np.concatenate((- np.flip(positive_positions_y), np.zeros(1),
                                                            positive_positions_y))
        position_array = np.array(np.meshgrid(position_array_x, position_array_y))

    #===========================================================================
    # Arrays
    final_coarse_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)
    final_wavelet_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)

    coarse_array_1 = np.zeros((image_x_size, image_y_size, n_levels), dtype = np.float32)
    coarse_array_2 = np.zeros((image_x_size, image_y_size, n_levels), dtype = np.float32)

    coarse_array_1[:,:,0] = np.copy(image)
    coarse_array_2[:,:,0] = np.copy(image)

    #===========================================================================
    # Convolution 1
    for level in range(0, n_levels - 1):

        position_array_lvl = np.array(position_array) * np.power(2, level)
        position_array_lvl = position_array_lvl.astype(np.int)

        if filter.ndim == 2:
            with np.nditer(coarse_array_1[:,:,level + 1], flags = ['multi_index'], op_flags = ['readwrite']) as it:
                for coeff in it:
                    position_array_coeff = np.copy(position_array_lvl)
                    position_array_coeff[0] += it.multi_index[0]
                    position_array_coeff[1] += it.multi_index[1]
                    position_array_coeff = boundary_conditions(position_array_coeff, interp_im.shape, 'loop')

                    coeff += np.sum(coarse_array_1[position_array_coeff[0], position_array_coeff[1], level] * filter)

        elif filter.ndim == 1:

            temporary_copy = np.copy(coarse_array_1[:,:,level + 1])
            # Over X axis

            for x in range(0, image_x_size ):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += x
                position_array_coeff = boundary_conditions(position_array_coeff, image.shape[0], 'loop')

                temporary_copy[x,:] = np.matmul(np.array([filter]), coarse_array_1[position_array_coeff, :, level])

            for y in range(0, image_y_size ):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += y
                position_array_coeff = boundary_conditions(position_array_coeff, image.shape[1] , 'loop')

                coarse_array_1[:,y,level + 1] = np.matmul(np.array([filter]), temporary_copy[ :, position_array_coeff].T)[0]

    # Convolution 2
    for level in range(0, n_levels - 1):

        position_array_lvl = np.array(position_array) * np.power(2, level + 0.5)
        position_array_lvl_ceil = np.ceil(position_array_lvl).astype(np.int)
        position_array_lvl_floor = np.floor(position_array_lvl).astype(np.int)

        if filter.ndim == 2:
            with np.nditer(coarse_array_2[:,:,level + 1], flags = ['multi_index'], op_flags = ['readwrite']) as it:
                for coeff in it:
                    position_array_coeff = np.copy(position_array_lvl)
                    position_array_coeff[0] += it.multi_index[0]
                    position_array_coeff[1] += it.multi_index[1]
                    position_array_coeff = boundary_conditions(position_array_coeff, interp_im.shape, 'loop')

                    coeff += np.sum(coarse_array_2[position_array_coeff[0], position_array_coeff[1], level] * filter)

        elif filter.ndim == 1:

            temporary_copy = np.copy(coarse_array_2[:,:,level])

            for x in range(0, image_x_size):

                position_array_ceil_coeff = np.copy(position_array_lvl_ceil)
                position_array_ceil_coeff += x
                position_array_ceil_coeff = boundary_conditions(position_array_ceil_coeff, image_x_size, 'loop')

                position_array_floor_coeff = np.copy(position_array_lvl_floor)
                position_array_floor_coeff += x
                position_array_floor_coeff = boundary_conditions(position_array_floor_coeff, image_x_size, 'loop')

                temporary_copy[x,:] = np.matmul(np.array([filter]),\
                            0.5 * (coarse_array_2[position_array_ceil_coeff, :, level]\
                            + coarse_array_2[position_array_floor_coeff, :, level]))

            for y in range(0, image_y_size):

                position_array_ceil_coeff = np.copy(position_array_lvl_ceil)
                position_array_ceil_coeff += y
                position_array_ceil_coeff = boundary_conditions(position_array_ceil_coeff, image_y_size, 'loop')

                position_array_floor_coeff = np.copy(position_array_lvl_floor)
                position_array_floor_coeff += y
                position_array_floor_coeff = boundary_conditions(position_array_floor_coeff, image_y_size, 'loop')

                #pdb.set_trace()

                coarse_array_2[:,y,level + 1] = np.matmul(np.array([filter]), \
                            0.5 * (temporary_copy[ :, position_array_ceil_coeff] \
                            + temporary_copy[ :, position_array_floor_coeff]).T)[0]

    voice = 0
    stds = []
    for level in range(0, n_levels - 1):

        final_wavelet_array[:,:,voice] = coarse_array_2[:,:,level] - coarse_array_1[:,:,level+1]
        final_coarse_array[:,:,voice] = coarse_array_2[:,:,level]
        print(level, voice, np.mean(final_wavelet_array[:,:,voice]), np.std(final_wavelet_array[:,:,voice]))
        stds.append(np.std(final_wavelet_array[:,:,voice]))
        voice += 1

        final_wavelet_array[:,:,voice] = coarse_array_1[:,:,level+1] - coarse_array_2[:,:,level+1]
        final_coarse_array[:,:,voice] = coarse_array_1[:,:,level+1]
        print(level, voice, np.mean(final_wavelet_array[:,:,voice]), np.std(final_wavelet_array[:,:,voice]))
        stds.append(np.std(final_wavelet_array[:,:,voice]))
        voice += 1

    final_wavelet_array[:,:,-1] = np.copy(coarse_array_2[:,:,-1])

    return final_coarse_array, final_wavelet_array, stds

def interlaced_atrous_filter(image, n_levels, n_voices, filter, verbose = False):

    '''
    Interlaced à trous algorithm, doc to do
    '''

    #===========================================================================
    # Exceptions
    if image.ndim != 2:
        raise DawisDimensionError('Not an image -->', image.ndim, 2)
    image_x_size = image.shape[0]
    image_y_size = image.shape[1]

    if filter.ndim > 2:
        raise DawisDimensionError('Not a 1D or 2D array -->', image.ndim, 2)

    if n_levels <= 0:
        raise DawisValueError('Number of wavelet planes cannot be negative -->', \
                                                n_levels, 'positive integer')

    #===========================================================================
    # Position array for filter
    if filter.ndim == 1:
        filter_ctr_index = filter.size // 2 + 1
        filter_half_size = filter.size // 2
        positive_positions = np.logspace( 0, filter_half_size - 1, \
                                            num = filter_half_size, base = 2)
        position_array = np.concatenate((- np.flip(positive_positions), np.zeros(1),
                                                            positive_positions))
    elif filter.ndim == 2:
        filter_ctr_index = (filter.shape[0] // 2 + 1, filter.shape[1] // 2 + 1 )
        filter_half_size = (filter.shape[0] // 2, filter.shape[1] // 2 )
        positive_positions_x = np.logspace( 0, filter_half_size[0] - 1, \
                                            num = filter_half_size[0], base = 2)
        positive_positions_y = np.logspace( 0, filter_half_size[1] - 1, \
                                                    num = filter_half_size[1], base = 2)
        position_array_x = np.concatenate((- np.flip(positive_positions_x), np.zeros(1),
                                                            positive_positions_x))
        position_array_y = np.concatenate((- np.flip(positive_positions_y), np.zeros(1),
                                                            positive_positions_y))
        position_array = np.array(np.meshgrid(position_array_x, position_array_y))

    #===========================================================================
    # Arrays
    final_coarse_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)
    final_wavelet_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)

    coarse_array_1 = np.zeros((image_x_size, image_y_size, n_levels), dtype = np.float32)
    coarse_array_2 = np.zeros((image_x_size, image_y_size, n_levels), dtype = np.float32)

    coarse_array_1[:,:,0] = np.copy(image)
    coarse_array_2[:,:,0] = np.copy(image)

    filter_test = np.array([ 0.0, 1/6, 2/3, 1/6, 0.0 ]) # from scipy.signal.bslpine with scale np.sqrt(2)

    #===========================================================================
    # Convolution 1
    for level in range(0, n_levels - 1):

        position_array_lvl = np.array(position_array) * np.power(2, level)
        position_array_lvl = position_array_lvl.astype(np.int)

        if filter.ndim == 2:
            with np.nditer(coarse_array_1[:,:,level + 1], flags = ['multi_index'], op_flags = ['readwrite']) as it:
                for coeff in it:
                    position_array_coeff = np.copy(position_array_lvl)
                    position_array_coeff[0] += it.multi_index[0]
                    position_array_coeff[1] += it.multi_index[1]
                    position_array_coeff = boundary_conditions(position_array_coeff, interp_im.shape, 'loop')

                    coeff += np.sum(coarse_array_1[position_array_coeff[0], position_array_coeff[1], level] * filter)

        elif filter.ndim == 1:

            temporary_copy = np.copy(coarse_array_1[:,:,level + 1])
            # Over X axis

            for x in range(0, image_x_size ):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += x
                position_array_coeff = boundary_conditions(position_array_coeff, image.shape[0], 'loop')

                temporary_copy[x,:] = np.matmul(np.array([filter]), coarse_array_1[position_array_coeff, :, level])

            for y in range(0, image_y_size ):

                position_array_coeff = np.copy(position_array_lvl)
                position_array_coeff += y
                position_array_coeff = boundary_conditions(position_array_coeff, image.shape[1] , 'loop')

                coarse_array_1[:,y,level + 1] = np.matmul(np.array([filter]), temporary_copy[ :, position_array_coeff].T)[0]

    # Convolution 2
    for level in range(0, n_levels - 1):

        position_array_lvl = np.array(position_array) * np.power(2, level)
        position_array_lvl_ceil = np.ceil(position_array_lvl).astype(np.int)
        position_array_lvl_floor = np.floor(position_array_lvl).astype(np.int)

        if filter.ndim == 2:
            with np.nditer(coarse_array_2[:,:,level + 1], flags = ['multi_index'], op_flags = ['readwrite']) as it:
                for coeff in it:
                    position_array_coeff = np.copy(position_array_lvl)
                    position_array_coeff[0] += it.multi_index[0]
                    position_array_coeff[1] += it.multi_index[1]
                    position_array_coeff = boundary_conditions(position_array_coeff, interp_im.shape, 'loop')

                    coeff += np.sum(coarse_array_2[position_array_coeff[0], position_array_coeff[1], level] * filter_test)

        elif filter.ndim == 1:

            temporary_copy = np.copy(coarse_array_2[:,:,level])

            for x in range(0, image_x_size):

                position_array_ceil_coeff = np.copy(position_array_lvl_ceil)
                position_array_ceil_coeff += x
                position_array_ceil_coeff = boundary_conditions(position_array_ceil_coeff, image_x_size, 'loop')

                position_array_floor_coeff = np.copy(position_array_lvl_floor)
                position_array_floor_coeff += x
                position_array_floor_coeff = boundary_conditions(position_array_floor_coeff, image_x_size, 'loop')

                temporary_copy[x,:] = np.matmul(np.array([filter_test]),\
                            0.5 * (coarse_array_2[position_array_ceil_coeff, :, level]\
                            + coarse_array_2[position_array_floor_coeff, :, level]))

            for y in range(0, image_y_size):

                position_array_ceil_coeff = np.copy(position_array_lvl_ceil)
                position_array_ceil_coeff += y
                position_array_ceil_coeff = boundary_conditions(position_array_ceil_coeff, image_y_size, 'loop')

                position_array_floor_coeff = np.copy(position_array_lvl_floor)
                position_array_floor_coeff += y
                position_array_floor_coeff = boundary_conditions(position_array_floor_coeff, image_y_size, 'loop')

                #pdb.set_trace()

                coarse_array_2[:,y,level + 1] = np.matmul(np.array([filter_test]), \
                            0.5 * (temporary_copy[ :, position_array_ceil_coeff] \
                            + temporary_copy[ :, position_array_floor_coeff]).T)[0]

    voice = 0
    stds = []
    for level in range(0, n_levels - 1):

        final_wavelet_array[:,:,voice] = coarse_array_1[:,:,level] - coarse_array_2[:,:,level+1]
        final_coarse_array[:,:,voice] = coarse_array_1[:,:,level]
        print(level, voice, np.mean(final_wavelet_array[:,:,voice]), np.std(final_wavelet_array[:,:,voice]))
        stds.append(np.std(final_coarse_array[:,:,voice]))
        voice += 1

        final_wavelet_array[:,:,voice] = coarse_array_2[:,:,level+1] - coarse_array_1[:,:,level+1]
        final_coarse_array[:,:,voice] = coarse_array_2[:,:,level+1]
        print(level, voice, np.mean(final_wavelet_array[:,:,voice]), np.std(final_wavelet_array[:,:,voice]))
        stds.append(np.std(final_coarse_array[:,:,voice]))
        voice += 1

    final_wavelet_array[:,:,-1] = np.copy(coarse_array_2[:,:,-1])

    return final_coarse_array, final_wavelet_array, stds

def interlaced_gaus(image, n_levels, n_voices, verbose = False):
    '''
    Interlaced à trous algorithm, doc to do
    '''
    #===========================================================================
    # Exceptions
    if image.ndim != 2:
        raise DawisDimensionError('Not an image -->', image.ndim, 2)
    image_x_size = image.shape[0]
    image_y_size = image.shape[1]

    #===========================================================================
    # Arrays
    final_coarse_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)
    final_wavelet_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)

    #===========================================================================
    # Convolution 0

    coarse_array_1 = np.zeros((image_x_size, image_y_size, n_levels), dtype = np.float32)
    coarse_array_2 = np.zeros((image_x_size, image_y_size, n_levels), dtype = np.float32)

    coarse_array_1[:,:,0] = np.copy(image)

    coarse_array_2[:,:,0] = gaussian_filter(image, sigma = 0.5 * np.sqrt(2))
    #coarse_array_2[:,:,0] = interp_im


    #===========================================================================
    # Convolution
    for level in range(0, n_levels - 1):
        coarse_array_1[:,:,level+1] = gaussian_filter(coarse_array_1[:,:,0], sigma = 2**(level+1))
        coarse_array_2[:,:,level+1] = gaussian_filter(coarse_array_2[:,:,0], sigma = 2**(level+1) * np.sqrt(2))


    voice = 0
    stds = []
    for level in range(0, n_levels - 1):

        final_wavelet_array[:,:,voice] = coarse_array_1[:,:,level] - coarse_array_2[:,:,level]
        final_coarse_array[:,:,voice] = coarse_array_1[:,:,level]
        print(level, voice, np.mean(final_wavelet_array[:,:,voice]), np.std(final_wavelet_array[:,:,voice]))
        stds.append(np.std(final_wavelet_array[:,:,voice]))
        voice += 1

        final_wavelet_array[:,:,voice] = coarse_array_2[:,:,level] - coarse_array_1[:,:,level+1]
        final_coarse_array[:,:,voice] = coarse_array_2[:,:,level]
        print(level, voice, np.mean(final_wavelet_array[:,:,voice]), np.std(final_wavelet_array[:,:,voice]))
        stds.append(np.std(final_wavelet_array[:,:,voice]))
        voice += 1

    final_wavelet_array[:,:,-1] = np.copy(coarse_array_2[:,:,-1])

    return final_coarse_array, final_wavelet_array, stds

def interlaced_spline(image, n_levels, n_voices, verbose = False):

    '''
    Interlaced à trous algorithm, doc to do
    '''
    #===========================================================================
    # Exceptions
    if image.ndim != 2:
        raise DawisDimensionError('Not an image -->', image.ndim, 2)
    image_x_size = image.shape[0]
    image_y_size = image.shape[1]

    #===========================================================================
    # Arrays
    final_coarse_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)
    final_wavelet_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)

    final_coarse_array[:,:,0] = np.copy(image)

    #===========================================================================
    # Convolution
    z = 0
    for level in np.arange(0, n_levels - 1, 0.5):

        x = np.arange( -1 * 2**level, 0.5 + 2**level, 1)

        bspl = bspline(x / 2**level, 3) / 2**level
        n = x.size
        bspl2d = np.zeros((n,n))
        for i in range(n):
            for j in range(n):
                bspl2d[i,j] = bspl[i] * bspl[j]

        #pdb.set_trace()
        final_coarse_array[:,:,z + 1] = convolve2d(final_coarse_array[:,:,z], bspl2d, 'same', 'wrap')

        z += 1

    #===========================================================================
    stds = []
    for level in range(0, 2 * n_levels - 1):

        final_wavelet_array[:,:,level] = final_coarse_array[:,:,level] - final_coarse_array[:,:,level+1]
        stds.append(np.std(final_wavelet_array[:,:,level]))

    final_wavelet_array[:,:,-1] = np.copy(final_coarse_array[:,:,-1])

    return final_coarse_array, final_wavelet_array, stds

def interlaced_mix_atrous_spline(image, n_levels, n_voices, verbose = False):

    '''
    Interlaced à trous algorithm, doc to do
    '''
    #===========================================================================
    # Exceptions
    if image.ndim != 2:
        raise DawisDimensionError('Not an image -->', image.ndim, 2)
    image_x_size = image.shape[0]
    image_y_size = image.shape[1]

    #===========================================================================
    # Arrays
    final_coarse_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)
    final_wavelet_array = np.zeros((image_x_size, image_y_size, 2 * n_levels), dtype = np.float32)

    final_coarse_array[:,:,0] = np.copy(image)

    #===========================================================================
    # Convolution
    z = 0
    for level in np.arange(0, n_levels - 1, 0.5):

        print(z)
        x = np.arange( -1 * 2**level, 0.5 + 2**level, 1)

        bspl = bspline(x / 2**level, 3) / 2**level
        n = x.size
        bspl2d = np.zeros((n,n))
        for i in range(n):
            for j in range(n):
                bspl2d[i,j] = bspl[i] * bspl[j]

        #pdb.set_trace()
        final_coarse_array[:,:,z + 1] = convolve2d(final_coarse_array[:,:,z], bspl2d, 'same', 'wrap')

        z += 1

    #===========================================================================
    stds = []
    for level in range(0, 2 * n_levels - 1):

        final_wavelet_array[:,:,level] = final_coarse_array[:,:,level] - final_coarse_array[:,:,level+1]
        print(level, np.mean(final_wavelet_array[:,:,level]), np.std(final_wavelet_array[:,:,level]))
        stds.append(np.std(final_wavelet_array[:,:,level]))

    final_wavelet_array[:,:,-1] = np.copy(final_coarse_array[:,:,-1])

    return final_coarse_array, final_wavelet_array, stds

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


if __name__ == '__main__':

    from numpy.random import normal
    from astropy.io import fits
    import sys
    sys.path.append('/home/ellien/DAWIS/codes/')
    from DL_reconstruct import DL_atrou

    im = normal(0, 1,(2048, 2048))
    #im = np.zeros((1000, 1000))
    #im[500,500] = 1

    #im = fits.getdata('/home/ellien/wavelets/MW_star_acs.fits')
    #im = fits.getdata('/home/ellien/wavelets/A1365.rebin.fits')
    #im = fits.getdata('/home/ellien/devd/tests/cat_gal_004_z_0.1_Ficl_0.4_Re_050_noise_megacam.fits')
    #nx, ny = np.shape(im)
    #anx, any = np.int(nx * np.sqrt(2) / 2), np.int(ny * np.sqrt(2) / 2)
    #im = congrid(im, (1024, 1024), method = 'spline')
    n_levels = 11

    # spline lissés
    startTime = datetime.now()
    bspl = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
    cbspl_new, wbspl_new = interlaced_atrous_congrid(image = im, n_levels = n_levels, filter = bspl, n_voices = 2)
    cbspl_new = np.delete(cbspl_new, obj = 1, axis = 2 ) # delete coarse scale 2^0.5
    wbspl_new = np.delete(wbspl_new, obj = 0, axis = 2 ) # delete coarse scale 2^0.5
    wbspl_new[:,:,0] = cbspl_new[:,:,0] - cbspl_new[:,:,1]
    print(datetime.now() - startTime)

    dcc_new = datacube(cbspl_new)
    dcc_new.waveplot(name = 'coarse', ncol = 7)
    dcw_new = datacube(wbspl_new)
    dcw_new.waveplot(name = 'wav', ncol = 7, cmap = 'hot' )

    # a trous
    bspl = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
    startTime = datetime.now()
    cbspl_old, wbspl_old = atrous(image = im, n_levels = n_levels, filter = bspl)
    stds_old = np.std(wbspl_old)
    print(datetime.now() - startTime)

    dcc_old = datacube(cbspl_old)
    #dcc_old.waveplot(name = 'coarse', ncol = 7, norm=matplotlib.colors.SymLogNorm(10**-5))
    dcw_old = datacube(wbspl_old)
    #dcw_old.waveplot(name = 'wav', ncol = 7, cmap = 'hot', norm=matplotlib.colors.SymLogNorm(10**-5) )

    #ovwav
    startTime = datetime.now()
    cbspl_ov = DL_atrou(im, n_levels, WAV_COEFF = False)
    wbspl_ov = DL_atrou(im, n_levels, WAV_COEFF = True)
    print(datetime.now() - startTime)

    # plots stds
    x_old = np.logspace(start = 0, stop = n_levels, num = n_levels + 1, base = 2)
    x_new = np.logspace(start = 0, stop = n_levels, num = 2 * n_levels + 1, base = 2)
    x_new = np.delete(x_new, 1)

    plt.figure()
    plt.title('stds of coarse planes')
    plt.plot(x_new[:-1], np.std(cbspl_new, axis = (0,1)), 'ro', linestyle = '-', label = 'interlaced à trous')
    plt.plot(x_old, np.std(cbspl_old, axis = (0,1)), 'bo', linestyle = '-', label = 'regular à trous')
    #plt.plot(x_old[:-1], np.std(cbspl_ov, axis = (0,1)), 'go', linestyle = '-')
    plt.xscale('log', basex = 2)
    plt.yscale('log')
    plt.legend()
    plt.show()

    plt.figure()
    plt.title('means of coarse planes')
    plt.plot(x_new[:-1], np.mean(cbspl_new, axis = (0,1)), 'ro', linestyle = '-', label = 'interlaced à trous')
    plt.plot(x_old, np.mean(cbspl_old, axis = (0,1)), 'bo', linestyle = '-', label = 'regular à trous')
    #plt.plot(x_old[:-1], np.mean(cbspl_ov, axis = (0,1)), 'go', linestyle = '-')
    plt.xscale('log', basex = 2)
    plt.legend()
    plt.show()

    plt.figure()
    plt.title('stds of wavelet planes')
    plt.plot(x_new[:-1], np.std(wbspl_new, axis = (0,1)), 'ro', linestyle = '-', label = 'interlaced à trous')
    plt.plot(x_old[:-1], np.std(wbspl_old, axis = (0,1)), 'bo', linestyle = '-', label = 'regular à trous')
    #plt.plot(x_old[:-1], np.std(wbspl_ov, axis = (0,1)), 'go', linestyle = '-')
    plt.xscale('log', basex = 2)
    plt.yscale('log')
    plt.legend()
    plt.show()

    plt.figure()
    plt.title('means of wavelet planes')
    plt.plot(x_new[:-1], np.mean(wbspl_new, axis = (0,1)), 'ro', linestyle = '-', label = 'interlaced à trous')
    plt.plot(x_old[:-1], np.mean(wbspl_old, axis = (0,1)), 'bo', linestyle = '-', label = 'regular à trous')
    #plt.plot(x_old[:-1], np.mean(wbspl_ov, axis = (0,1)), 'go', linestyle = '-')
    plt.xscale('log', basex = 2)
    plt.legend()
    plt.show()

    print(np.std(wbspl_new, axis = (0,1)))

    #bspl = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
    #startTime = datetime.now()
    #cbspl, wbspl = atrous(image = im, n_levels = 6, filter = bspl)

    #wp_new_check = np.zeros(wp_old_check.shape)
    #for level in range(0, n_levels - 1):
    #   wp_new_check[:,:,level] = dcc.array[:,:,2*level] - dcc.array[:,:,2*(level+1)]

    #plt.figure()
    #plt.title('diff new/old a trous')
    #plt.imshow( 100 * (wp_new_check[:,:,0] - wp_old_check[:,:,0]) / wp_new_check[:,:,0], vmax = 1, vmin = -1 )
    #plt.show()

    # print(im.shape)
    # starlet = 1 / 256. * np.array([ [1,  4,  6,  4, 1],
    #                                 [4, 16, 24, 16, 4],
    #                                 [6, 24, 36, 24, 6],
    #                                 [4, 16, 24, 16, 4],
    #                                 [1,  4,  6,  4, 1] ])
    # startTime = datetime.now()
    # cstar, wstar = atrous(image = im, n_levels = 6, filter = starlet)
    # print(datetime.now() - startTime)

    #bspl = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
    #startTime = datetime.now()
    #cbspl, wbspl = atrous(image = im, n_levels = 6, filter = bspl)
    #print(datetime.now() - startTime)

    # haar = np.array([ 0.5, 0, 0.5 ])
    # startTime = datetime.now()
    # chaar, whaar = atrous(image = im, n_levels = 6, filter = haar)
    # print(datetime.now() - startTime)

    # diff_bspl_starlet = datacube(wstar - wbspl)
    # diff_bspl_starlet.waveplot(name = 'Diff btw starlet and bspl', save_path = '/home/ellien/devd/tests/diff_bspl_starlet.pdf')

    # diff_bspl_haar = datacube(whaar - wbspl)
    # diff_bspl_haar.waveplot(name = 'Diff btw haar and bspl', save_path = '/home/ellien/devd/tests/diff_bspl_haar.pdf')



import numpy as n
import scipy.interpolate
import scipy.ndimage

def congrid(a, newdims, method='linear', centre=False, minusone=False):
    '''Arbitrary resampling of source array to new dimension sizes.
    Currently only supports maintaining the same number of dimensions.
    To use 1-D arrays, first promote them to shape (x,1).

    Uses the same parameters and creates the same co-ordinate lookup points
    as IDL''s congrid routine, which apparently originally came from a VAX/VMS
    routine of the same name.

    method:
    neighbour - closest value from original data
    nearest and linear - uses n x 1-D interpolations using
                         scipy.interpolate.interp1d
    (see Numerical Recipes for validity of use of n 1-D interpolations)
    spline - uses ndimage.map_coordinates

    centre:
    True - interpolation points are at the centres of the bins
    False - points are at the front edge of the bin

    minusone:
    For example- inarray.shape = (i,j) & new dimensions = (x,y)
    False - inarray is resampled by factors of (i/x) * (j/y)
    True - inarray is resampled by(i-1)/(x-1) * (j-1)/(y-1)
    This prevents extrapolation one element beyond bounds of input array.
    '''
    if not a.dtype in [n.float64, n.float32]:
        a = n.cast[float](a)

    m1 = n.cast[int](minusone)
    ofs = n.cast[int](centre) * 0.5
    old = n.array( a.shape )
    ndims = len( a.shape )
    if len( newdims ) != ndims:
        print("[congrid] dimensions error. " \
              "This routine currently only support " \
              "rebinning to the same number of dimensions.")
        return None
    newdims = n.asarray( newdims, dtype=float )
    dimlist = []

    if method == 'neighbour':
        for i in range( ndims ):
            base = n.indices(newdims)[i]
            dimlist.append( (old[i] - m1) / (newdims[i] - m1) \
                            * (base + ofs) - ofs )
        cd = n.array( dimlist ).round().astype(int)
        newa = a[list( cd )]
        return newa

    elif method in ['nearest','linear']:
        # calculate new dims
        for i in range( ndims ):
            base = n.arange( newdims[i] )
            dimlist.append( (old[i] - m1) / (newdims[i] - m1) \
                            * (base + ofs) - ofs )
        # specify old dims
        olddims = [n.arange(i, dtype = n.float) for i in list( a.shape )]

        # first interpolation - for ndims = any
        mint = scipy.interpolate.interp1d( olddims[-1], a, kind=method )
        newa = mint( dimlist[-1] )
        trorder = [ndims - 1] + list( range( ndims - 1 ) )

        for i in range( ndims - 2, -1, -1 ):
            newa = newa.transpose( trorder )

            mint = scipy.interpolate.interp1d( olddims[i], newa, kind=method )
            newa = mint( dimlist[i] )

        if ndims > 1:
            # need one more transpose to return to original dimensions
            newa = newa.transpose( trorder )

        return newa
    elif method in ['spline']:
        oslices = [ slice(0,j) for j in old ]
        oldcoords = n.ogrid[oslices]
        nslices = [ slice(0,j) for j in list(newdims) ]
        newcoords = n.mgrid[nslices]

        newcoords_dims = list(range(n.ndim(newcoords)))
        #make first index last
        newcoords_dims.append(newcoords_dims.pop(0))
        newcoords_tr = newcoords.transpose(newcoords_dims)
        # makes a view that affects newcoords

        newcoords_tr += ofs

        deltas = (n.asarray(old) - m1) / (newdims - m1)
        newcoords_tr *= deltas

        newcoords_tr -= ofs

        newa = scipy.ndimage.map_coordinates(a, newcoords)
        return newa
    else:
        print("Congrid error: Unrecognized interpolation type.\n", \
              "Currently only \'neighbour\', \'nearest\',\'linear\',", \
              "and \'spline\' are supported.")
        return None



import os
import dawis

def test():

    indir = os.path.join(dawis.__path__[0], 'gallery')
    infile = 'tidal_group_f814w_sci.fits'
    outdir = input('DAWIS - In which directory should I write ?\n')
    n_cpus = int(input('DAWIS - How many CPUs available ?\n'))

    tau = 0.1   # Relative Threshold
    gamma = 0.5   # Attenuation factor
    ceps = 1E-4    # Convergence value for epsilon
    n_levels = 9    # Number of wavelet scales
    min_span = 2    # Minimum of wavelet scales spanned by an interscale tree (must be >= 1)
    max_span = 3    # Maximum number of wavelet scales spanned by an interscale tree
    lvl_sep_big = 5     # Scale at wich mix_span & max_span are set to 1
    extent_sep = 0.1    # Ratio n_pix/vignet under which the Haar wavelet is used for restoration
    lvl_sep_lin = 2     # Wavelet scale under which the Haar wavelet can be used for restoration
    max_iter = 500      # Maximum number of iterations
    data_dump = True    # Write data at each iteration /!\ demands lot of space on hardware /!\
    gif = True      # Make gifs of the run (need data_dump = True)
    starting_level = 2 # Starting wavelet scale (this is the third scale - Python convention 0 1 2)

    dawis.synthesis_by_analysis( indir = indir, infile = infile, outdir = outdir, n_cpus = n_cpus, n_levels = n_levels, \
                                        tau = tau, gamma = gamma, ceps = ceps, min_span = min_span, \
                                        max_span = max_span, lvl_sep_big = lvl_sep_big, max_iter = max_iter, \
                                        data_dump = data_dump, gif = gif )



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# This file is part of DAWIS.
#
# NAME : make_interscale_trees.py
# AUTHOR : Amael Ellien
# LAST MODIFICATION : 07/2021
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

from datetime import datetime
from dawis.atrous import atrous
import pdb
import matplotlib.pyplot as plt
import pickle
import numpy as np
import logging
import ultranest
import ray
from matplotlib.colors import SymLogNorm
from skimage.feature import peak_local_max
from scipy import ndimage as ndi
from skimage.segmentation import watershed
from skimage.measure import label, regionprops

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

class interscale_tree(object):
    """docstring for ."""

    def __init__(self, interscale_maximum, region_list, wavelet_datacube, label_datacube, clip = 30):

        self.region_list = region_list
        self.interscale_maximum = interscale_maximum
        self.n_regions = len(region_list)
        self.span_levels = np.size(np.unique([ x.level for x in region_list ]))
        self.clip = clip
        self.eccentricity = self.interscale_maximum.eccentricity

        x_min = np.min([ x.bbox[0] for x in self.region_list ])
        y_min = np.min([ x.bbox[1] for x in self.region_list ])
        x_max = np.max([ x.bbox[2] for x in self.region_list ])
        y_max = np.max([ x.bbox[3] for x in self.region_list ])

        # Extent
        clip = np.copy(label_datacube.array[ x_min:x_max, \
                                             y_min:y_max, \
                                             : ])
        self.bbox = (x_min, y_min, x_max, y_max)
        extent_im = np.sum( self.tubes( wavelet_datacube, label_datacube )[1], axis = 2 )
        n_pix = np.size( np.where( extent_im != 0. )[0] )
        self.extent = n_pix / ( ( x_max - x_min ) * ( y_max - y_min ) )

        # Make vignet larger
        x_min -= np.int( ( x_max - x_min ) / 2. )
        x_max += np.int( ( x_max - x_min ) / 2. )
        if x_min < 0: x_min = 0
        if x_max > label_datacube.x_size: x_max = label_datacube.x_size

        y_min -= np.int( ( y_max - y_min ) / 2. )
        y_max += np.int( ( y_max - y_min ) / 2. )
        if y_min < 0: y_min = 0
        if y_max > label_datacube.y_size: y_max = label_datacube.y_size

        # Clip
        if x_max - x_min < self.clip:
            x_min = int( (x_max + x_min) / 2. - self.clip / 2. + 1.)   # we resize it to the minclip
            x_max = int( (x_max + x_min) / 2. + self.clip / 2. + 1.)   # value.
            if x_min < 0: x_min = 0
            if x_max > label_datacube.x_size: x_max = label_datacube.x_size

        if y_max - y_min < self.clip:
            y_min = int( (y_max + y_min) / 2. - self.clip / 2. + 1.)   # we resize it to the minclip
            y_max = int( (y_max + y_min) / 2. + self.clip / 2. + 1.)   # value.
            if y_min < 0: y_min = 0
            if y_max > label_datacube.y_size: y_max = label_datacube.y_size

        self.bbox = (x_min, y_min, x_max, y_max)
        self.x_size = x_max - x_min
        self.y_size = y_max - y_min

        # Bayesian parameter bounds
        self.lowb = 0.
        self.highb = 2.


    def plot(self, wavelet_datacube, label_datacube, name = None, show = True, save_path = None, **kwargs):

        fig, ax = plt.subplots(1, 2, sharex = True, sharey = True, \
                                  gridspec_kw = {'hspace': 0.5, 'wspace': 0.5})

        if name == None:
            fig.suptitle('Interscale tree of region n°%d' %(self.interscale_maximum.label) )
        else:
            fig.suptitle(name)

        wavelet_image = np.sum(wavelet_datacube.array[ self.bbox[0]:self.bbox[2], \
                                                       self.bbox[1]:self.bbox[3], \
                                                       0:self.interscale_maximum.level + 1 ], axis = 2)

        try:
            norm = kwargs.pop('norm')
            im = ax[0].imshow( wavelet_image, \
                    extent = ( self.bbox[1], self.bbox[3], \
                               self.bbox[0], self.bbox[2] ), \
                    origin = 'lower', \
                    **kwargs )
            plt.colorbar(im, ax = ax[0], orientation = 'vertical')
        except:
            logthresh = 1E-3
            norm = SymLogNorm(logthresh)
            im = ax[0].imshow( wavelet_image, extent = ( self.bbox[1], self.bbox[3], \
                                                         self.bbox[0], self.bbox[2] ), \
                                              origin = 'lower', \
                                              norm = norm, \
                                              **kwargs )

            maxlog = np.log10( np.ceil( np.max( wavelet_image )))
            minlog = np.log10( np.ceil( - np.min( wavelet_image )))
            tick_locations = (list(- np.logspace(1, minlog, num = 2, base = 10)) \
                                + [0] \
                                + list(np.logspace(1, maxlog, num = 2, base = 10)))
            plt.colorbar(im, ax = ax[0], orientation = 'vertical', \
                                         ticks = tick_locations,
                                         aspect = 10,
                                         shrink = 0.5)


        label_image = np.zeros((self.bbox[2] - self.bbox[0], \
                                self.bbox[3] - self.bbox[1]))

        for i, region in enumerate(self.region_list):

            ax[0].plot( region.y_max, region.x_max, '+' )

            clip = np.copy(label_datacube.array[ self.bbox[0]:self.bbox[2], \
                                                 self.bbox[1]:self.bbox[3], \
                                                 np.int(region.level) ])

            clip[np.where(clip != region.label)] = 0.

            try:
                ax[0].contour( clip, region.label,
                                     extent = ( self.bbox[1], self.bbox[3], \
                                                self.bbox[0], self.bbox[2] ), \
                                     levels = 0, \
                                     color = 'cyan')
            except:
                print('Region too small for contours.')

            clip[np.where(clip == region.label)] = 1
            label_image += clip
            ax[1].plot( region.y_max, region.x_max, '+' )

        ax[1].imshow(label_image, extent = ( self.bbox[1], self.bbox[3], \
                                             self.bbox[0], self.bbox[2] ), \
                                  origin = 'lower',
                                  cmap = 'binary'  )


        if save_path != None:
            plt.savefig(save_path, format = 'pdf')

        if show == True:
            plt.show()
        else:
            plt.close()

    def tubes( self, wavelet_datacube, label_datacube ):

        label_tube = np.zeros((self.bbox[2] - self.bbox[0], \
                               self.bbox[3] - self.bbox[1], \
                               self.interscale_maximum.level + 1))

        wavelet_tube = np.copy(label_tube)

        for region in self.region_list:

            wavelet_tube[ region.coords[:,0] - self.bbox[0], \
                          region.coords[:,1] - self.bbox[1], \
                          region.level] = wavelet_datacube.array[ region.coords[:,0], \
                                                          region.coords[:,1], \
                                                          region.level ]

            label_tube[ region.coords[:,0] - self.bbox[0], \
                        region.coords[:,1] - self.bbox[1], \
                        region.level] = label_datacube.array[ region.coords[:,0], \
                                                              region.coords[:,1], \
                                                              region.level ]

        support_tube = np.copy(label_tube)
        support_tube[support_tube != 0] = 1.

        return wavelet_tube, support_tube, label_tube

    def synthesis_adjoint( self, wavelet_tube, filter ):

        '''
        Tild operator Â associated to the wavelet transform WT. It is given by :

                        Â(W)=SUM_i(H(i)*...*H(N-1)*W(i)) with W a wavelet object
                                                       and H a low-band filter.

        The result is an approximated first reconstruction of the image F with W=WT(F). Here H is given by the 'a trou' algorithm in its smooth only version.
        doc to modif
        '''

        smooth_coeff = np.zeros(wavelet_tube.shape)

        for level in range(wavelet_tube.shape[2]):

            smooth_coeff[:, :, level] = atrous( wavelet_tube[:, :, level], \
                                               n_levels = level + 1, \
                                               filter = filter, \
                                               conditions = 'prolongation' )[0][:, :, level]
        return np.sum( smooth_coeff, axis = 2 )

    def synthesis_sum( self, wavelet_tube ):
        return np.sum( wavelet_tube, axis = 2 )

    def set_priors( self, cube ):
        '''WIP
        '''
        params = cube.copy()

        # All parameters have same unifor prios
        for i, par in enumerate( params ):
            par = cube[i] * (self.highb - self.lowb) + self.lowb

        return params

    def set_param_names( self, wavelet_tube ):
        '''WIP
        '''
        n_levels = wavelet_tube.shape[2]

        param_names = []
        for i in range( n_levels ):
            param_names.append( 'a%s' %(str(i)) )

        for i in range( n_levels ):
            param_names.append( 'b%s' %(str(i)) )

        return param_names

    def synthesis_likelihood( self, params ):
        '''WIP
        '''
        n_levels = wt.shape[2]

        smooth_coeff = np.zeros(wt.shape)
        adjt = np.copy(smooth_coeff[:,:,0])

        # Â(W) = SUM_i a_i * ( b_i * H(i) * ... * H(N - 1) * W(i))
        for level in range(n_levels):
            smooth_coeff[:, :, level] = atrous( wt[:, :, level], \
                                                n_levels = level + 1, \
                                                filter = bspl * params[ level + n_levels ], \
                                                conditions = 'prolongation' )[0][:, :, level]

        for level in range(n_levels):
            adjt +=  params[ level ] * smooth_coeff[:, :, level]

        r = adjt - im[self.bbox[0]:self.bbox[2], self.bbox[1]:self.bbox[3]]
        #r[r < 0] = 0
        covr = np.cov(r)
        pdb.set_trace()

        loglike = -0.5 * np.sum( r * np.linalg.inv(covr) )

        wr = wt - atrous(r, n_levels = wavelet_tube.shape[2], \
                                      filter = filter, \
                                      conditions = 'prolongation' )[1] * support_tube


        for level in range(n_levels):
            covwr = np.cov(wr[:,:,level])
            loglike += -0.5 * np.sum( wr[:,:,level] * np.linalg.inv(covwr) )

        return loglike

    def bayesian_minimization( self, original_image, wavelet_datacube, label_datacube, filter, verbose = False ):
        '''WIP
        '''
        wavelet_tube, support_tube, label_tube = self.tubes(wavelet_datacube, label_datacube)
        param_names = self.set_param_names( wavelet_tube )

        sampler = ultranest.ReactiveNestedSampler(param_names, \
                                                  self.synthesis_likelihood, \
                                                  self.set_priors,\
                                                  log_dir = None )
        result = sampler.run()
        sampler.print_results()

        return None

    def SD_minimization( self, wavelet_datacube, label_datacube, filter, \
                            synthesis_operator = 'ADJOINT', sigma_flux = 1E-3, \
                            max_iter = 200, verbose = False):
        '''
        simple gradient or steepest descent (bijaoui & rué, 1995).
        '''

        wavelet_tube, support_tube, label_tube = self.tubes(wavelet_datacube, label_datacube)

        if synthesis_operator == 'ADJOINT':
            reconstructed = self.synthesis_adjoint(wavelet_tube, filter = filter)
        elif synthesis_operator == 'SUM':
            reconstructed = self.synthesis_sum(wavelet_tube)

        reconstructed[ reconstructed < 0.0 ] = 0.0

        iter = 0
        total_flux = 1.0
        while iter <= max_iter :

            iter += 1
            wr = wavelet_tube - atrous(reconstructed, n_levels = wavelet_tube.shape[2], \
                                                      filter = filter, \
                                                      conditions = 'prolongation' )[1] * support_tube

            if synthesis_operator == 'ADJOINT':
                fr = self.synthesis_adjoint(wr, filter = filter )
            elif synthesis_operator == 'SUM':
                fr = self.synthesis_sum(wr)

            norm_wr = np.sqrt( np.sum( wr**2 ) )
            if norm_wr == 0:
                norm_wr = 1.

            alpha = np.sqrt( np.sum( fr**2 ) ) / norm_wr
            reconstructed = reconstructed + alpha * fr
            reconstructed[ reconstructed < 0.0 ] = 0.0

            old_total_flux = total_flux
            total_flux = np.sum(reconstructed)
            if old_total_flux == 0.:
                old_total_flux = 1.

            diff_flux = abs(total_flux - old_total_flux) / old_total_flux

            if ( diff_flux < sigma_flux ):
                if verbose == True:
                    print( "flux convergence", np.max(reconstructed), np.shape(reconstructed) )
                break

        wr = wavelet_tube - atrous(reconstructed, n_levels = wavelet_tube.shape[2], \
                                                      filter = filter, \
                                                      conditions = 'prolongation' )[1] * support_tube

        norm_wr = np.sqrt( np.sum( wr**2 ) )
        sum_wr = np.sum( wr )

        return reconstructed, sum_wr, norm_wr

    def CG_minimization(self, wavelet_datacube, label_datacube, filter, \
                            synthesis_operator = 'ADJOINT', step_size = 'FR', \
                            sigma_flux = 1E-3, max_iter = 200, verbose = False):
        '''
        conjugate gradient (modified from starck, murtagh & bijaoui, 1998).
        '''

        wavelet_tube, support_tube, label_tube = self.tubes(wavelet_datacube, label_datacube)
        n_levels = wavelet_tube.shape[2]

        if synthesis_operator == 'ADJOINT':
            reconstructed = self.synthesis_adjoint(wavelet_tube, filter = filter)
        elif synthesis_operator == 'SUM':
            reconstructed = self.synthesis_sum(wavelet_tube)

        wr = wavelet_tube - atrous(reconstructed, n_levels = n_levels, \
                                                  filter = filter, \
                                                  conditions = 'prolongation' )[1] * support_tube

        if synthesis_operator == 'ADJOINT':
            fr = self.synthesis_adjoint(wr, filter = filter )
        elif synthesis_operator == 'SUM':
            fr = self.synthesis_sum(wr)

        v = np.copy(fr)

        iter = 0
        total_flux = 1.0
        while iter <= max_iter:

            iter = iter + 1

            norm_wr = np.linalg.norm(wr)
            if norm_wr == 0:
                norm_wr = 1
            delta = np.linalg.norm(fr) / norm_wr
            reconstructed = reconstructed + delta * v
            reconstructed[ reconstructed < 0.0 ] = 0.0

            old_total_flux = total_flux
            total_flux = np.sum(reconstructed)

            if old_total_flux == 0.0:
                old_total_flux = 1.0

            diff_flux = abs( total_flux - old_total_flux ) / old_total_flux
            if diff_flux < sigma_flux:
                if verbose == True:
                    print( "flux convergence", np.max(reconstructed), np.shape(reconstructed) )
                break

            else:

                old_wr = np.copy(wr)
                old_fr = np.copy(fr)

                wr = wavelet_tube - atrous( reconstructed, n_levels = n_levels, \
                                                filter = filter, \
                                                conditions = 'prolongation' )[1] * support_tube

                if synthesis_operator == 'ADJOINT':
                    fr = self.synthesis_adjoint(wr, filter = filter )
                elif synthesis_operator == 'SUM':
                    fr = self.synthesis_sum(wr)

                squared_norm_old_fr = np.linalg.norm( old_fr )**2
                if squared_norm_old_fr == 0.:
                    squared_norm_old_fr = 1.0

                if step_size == 'FR':
                    beta = np.linalg.norm( fr )**2 / squared_norm_old_fr
                elif step_size == 'PRP':
                        beta = np.sum( (fr - old_fr) * fr ) / np.sum( fr * fr )
                elif step_size == 'HS':
                        beta = np.sum( (fr - old_fr) * fr ) / np.sum( (fr - old_fr) * v )
                elif step_size == 'CD':
                        beta = np.linalg.norm( fr )**2 / np.sum( v * old_fr )
                v = fr + beta * old_fr

        wr = wavelet_tube - atrous(reconstructed, n_levels = wavelet_tube.shape[2], \
                                                      filter = filter, \
                                                      conditions = 'prolongation' )[1] * support_tube

        norm_wr = np.sqrt( np.sum( wr**2 ) )
        sum_wr = np.sum( wr )

        return reconstructed, sum_wr, norm_wr

def write_interscale_trees_to_pickle(interscale_tree_list, filename, overwrite = True):

    if overwrite == True:
        mode = 'wb'
    else:
        mode = 'ab'

    with open(filename, mode) as outfile:
        pickle.dump(interscale_tree_list, outfile)

def read_interscale_trees_from_pickle(filename):

    with open(filename, 'rb') as infile:
        interscale_tree_list = pickle.load(infile)

    return interscale_tree_list

def enforce_monomodality(interscale_maximum, wavelet_datacube, label_datacube):

    # Label clip of interscale maximum region
    label_clip = np.copy( label_datacube.array[ :, :, interscale_maximum.level ])
    label_clip[ np.where(label_clip != interscale_maximum.label) ] = 0.

    # Wavelet clip of interscale maximum region
    wavelet_clip = np.copy( wavelet_datacube.array[ :, :, interscale_maximum.level ])
    wavelet_clip[ np.where( label_clip == 0 ) ] = 0.

    # Find secondary local wavelet maxima
    local_maxima_coo = peak_local_max( wavelet_clip, exclude_border = False )
    if len(local_maxima_coo) > 1:

        # Make mask image of local wavelet maxima
        local_maxima_clip = np.zeros_like( label_clip )
        local_maxima_clip[ tuple(local_maxima_coo.T) ] = 1

        # Create new labels for each local maxima, called markers
        marker_clip = label(local_maxima_clip)

        # Watershed algorithm to segment the interscale maximum into different regions
        # corresponding to local maxima
        segmented_label_clip = watershed( - wavelet_clip, markers = marker_clip, mask = label_clip )

        # Find which local maximum is interscale maximum
        local_label_max = segmented_label_clip[ interscale_maximum.x_max, interscale_maximum.y_max ]

        # Update global label datacube by setting to zero pixels not covered by
        # new region
        updated_labels = np.copy(label_datacube.array[ :, :, interscale_maximum.level ])
        updated_labels[  np.where( (label_datacube.array[ :, :, interscale_maximum.level ] == interscale_maximum.label) \
                                            & (segmented_label_clip != local_label_max) )] = 0.

        label_datacube.array[ :, :, interscale_maximum.level ] = updated_labels

        # Measure new region properties
        segmented_label_clip[ segmented_label_clip != local_label_max ] = 0.
        props_skimage = regionprops(segmented_label_clip, wavelet_clip)

        # Update interscale maximum properties that were modified by segmentation
        interscale_maximum.area = props_skimage[0].area
        interscale_maximum.bbox = props_skimage[0].bbox
        interscale_maximum.bbox_area = props_skimage[0].bbox_area
        interscale_maximum.centroid = props_skimage[0].centroid
        interscale_maximum.local_centroid = props_skimage[0].local_centroid
        interscale_maximum.coords = props_skimage[0].coords
        interscale_maximum.extent = props_skimage[0].extent
        interscale_maximum.eccentricity = props_skimage[0].eccentricity
        interscale_maximum.max_intensity = props_skimage[0].max_intensity
        interscale_maximum.min_intensity = props_skimage[0].min_intensity

    return interscale_maximum, label_datacube

@ray.remote
def interscale_connectivity_para( interscale_maximum_list, region_list, wavelet_datacube, label_datacube, level_maximum, min_span = 3, max_span = 3, lvl_sep_big = 6, monomodality = False, min_reg_size = 4, verbose = False):

    interscale_tree_list = []
    n_rejected = 0
    for interscale_maximum in interscale_maximum_list:

        if interscale_maximum.level >= lvl_sep_big :
            pmax_span = 1
            pmin_span = 1

        elif interscale_maximum.level < lvl_sep_big :
            pmax_span = max_span
            pmin_span = min_span

        tube = label_datacube.array[interscale_maximum.coords[:,0], interscale_maximum.coords[:,1], :level_maximum + 1]
        covered_label_list = list(np.unique(tube))
        covered_region_list = list(filter( lambda x: x.label in covered_label_list, region_list ))

        connected_region_list = list(filter( lambda x: (x.label in covered_label_list) & \
                                                       ( [x.x_max, x.y_max] in interscale_maximum.coords ) & \
                                                       ( x.area >= min_reg_size) & \
                                                       ( interscale_maximum.level - x.level <= pmax_span - 1 ), region_list ))

        span_levels = np.size(np.unique([ x.level for x in connected_region_list ]))

        if span_levels < pmin_span :
            #log.info('Rejected : level = %d span = %d' %(interscale_maximum.level, span_levels))
            n_rejected += 1
            continue

        interscale_tree_list.append(interscale_tree(interscale_maximum, connected_region_list, wavelet_datacube, label_datacube))

    return interscale_tree_list

def interscale_connectivity_serial( interscale_maximum_list, region_list, wavelet_datacube, label_datacube, level_maximum, min_span = 3, max_span = 3, lvl_sep_big = 6, monomodality = False, min_reg_size = 4, verbose = False):

    interscale_tree_list = []
    n_rejected = 0
    for interscale_maximum in interscale_maximum_list:

        if interscale_maximum.level >= lvl_sep_big :
            pmax_span = 1
            pmin_span = 1

        elif interscale_maximum.level < lvl_sep_big :
            pmax_span = max_span
            pmin_span = min_span

        tube = label_datacube.array[interscale_maximum.coords[:,0], interscale_maximum.coords[:,1], :level_maximum + 1]
        covered_label_list = list(np.unique(tube))
        covered_region_list = list(filter( lambda x: x.label in covered_label_list, region_list ))

        connected_region_list = list(filter( lambda x: (x.label in covered_label_list) & \
                                                       ( [x.x_max, x.y_max] in interscale_maximum.coords ) & \
                                                       ( x.area >= min_reg_size) & \
                                                       ( interscale_maximum.level - x.level <= pmax_span - 1 ), region_list ))

        span_levels = np.size(np.unique([ x.level for x in connected_region_list ]))

        if span_levels < pmin_span :
            #log.info('Rejected : level = %d span = %d' %(interscale_maximum.level, span_levels))
            n_rejected += 1
            continue

        interscale_tree_list.append(interscale_tree(interscale_maximum, connected_region_list, wavelet_datacube, label_datacube))

    return interscale_tree_list


def make_interscale_trees(region_list, wavelet_datacube, label_datacube, tau = 0.8, min_span = 3, max_span = 3, lvl_sep_big = 6, monomodality = False, min_reg_size = 4, size_patch = 100, n_cpus = 1, verbose = False):

    interscale_tree_list = []

    region_list.sort(key = lambda x: x.norm_max_intensity, reverse = True)

    i = 0
    levels_rejected = [ 0, label_datacube.z_size ]
    threshold_maximum = region_list[0].norm_max_intensity * tau
    level_maximum = region_list[0].level
    while region_list[i].level in levels_rejected :
        i += 1
        threshold_maximum = region_list[i].norm_max_intensity * tau
        level_maximum = region_list[i].level

    interscale_maximum_list = list(filter( lambda x: ( x.norm_max_intensity >= tau * threshold_maximum ) & \
                                                     ( x.level == level_maximum ) & \
                                                     ( x.area >= min_reg_size ), region_list))
    if verbose == True:
        log = logging.getLogger(__name__)
        log.info('Estimating global interscale maxima: %d found.' %(len(interscale_maximum_list)))


    for interscale_maximum in interscale_maximum_list:
        #if interscale_maximum.level >= lvl_sep_big :
        if monomodality == True:
            interscale_maximum, label_datacube = enforce_monomodality( interscale_maximum, wavelet_datacube, label_datacube )

    if (len(interscale_maximum_list) <= size_patch) or (n_cpus == 1):

        interscale_tree_list = interscale_connectivity_serial( interscale_maximum_list = interscale_maximum_list,  \
                                                               region_list = region_list,  \
                                                               wavelet_datacube = wavelet_datacube,  \
                                                               label_datacube = label_datacube,  \
                                                               level_maximum = level_maximum,  \
                                                               min_span = min_span,  \
                                                               max_span = max_span,  \
                                                               lvl_sep_big = lvl_sep_big, \
                                                               monomodality = monomodality, \
                                                               min_reg_size = min_reg_size, \
                                                               verbose = verbose )

    else:

        logging.info('Size interscale maximum patch (%d) greater than %d, activating Ray store.'%( len(interscale_maximum_list), size_patch ))
        ray.init(num_cpus = n_cpus)
        id_rl = ray.put(region_list)
        id_wdc = ray.put(wavelet_datacube)
        id_ldc = ray.put(label_datacube)
        id_level_maximum = ray.put(level_maximum)
        id_min_span = ray.put(min_span)
        id_max_span = ray.put(max_span)
        id_lvl_sep_big = ray.put(lvl_sep_big)
        id_min_reg_size = ray.put(min_reg_size)
        id_verbose = ray.put(verbose)
        id_monomodality = ray.put(monomodality)

        interscale_maximum_patch = []
        interscale_tree_patch = []

        for interscale_maximum in interscale_maximum_list:

            interscale_maximum_patch.append(interscale_maximum)
            if len(interscale_maximum_patch) >= size_patch:
                interscale_tree_patch.append( interscale_connectivity_para.remote( interscale_maximum_list = interscale_maximum_patch,  \
                                                                                   region_list = id_rl,  \
                                                                                   wavelet_datacube = id_wdc,  \
                                                                                   label_datacube = id_ldc,  \
                                                                                   level_maximum = id_level_maximum,  \
                                                                                   min_span = id_min_span,  \
                                                                                   max_span = id_max_span,  \
                                                                                   lvl_sep_big = id_lvl_sep_big, \
                                                                                   monomodality = id_monomodality, \
                                                                                   min_reg_size = id_min_reg_size, \
                                                                                   verbose = id_verbose ) )
                interscale_maximum_patch = []

        for id_patch in interscale_tree_patch:
            patch = ray.get( id_patch )
            interscale_tree_list = interscale_tree_list + patch

    ray.shutdown()

    if verbose == True:
        log.info('%d rejected.' %(len(interscale_maximum_list) - len(interscale_tree_list)))
    if not interscale_tree_list:
        log.info("No interscale maximum found. Please consider lowering parameters 'tau' or 'min_span' if this keeps happening at every level. ")

    return interscale_tree_list, label_datacube

if __name__ == '__main__':

    from make_regions import *
    from datacube import *
    from matplotlib.colors import SymLogNorm, LogNorm
    from astropy.io import fits
    from atrous import atrous
    from wavelet_transforms import bspl_atrous, wavelet_datacube
    from hard_threshold import hard_threshold
    from label_regions import label_regions
    from anscombe_transforms import *
    from skimage.metrics import structural_similarity as ssim

    hdu = fits.open('/home/ellien/devd/gallery/A1365.rebin.fits')
    im = hdu[0].data
    header = hdu[0].header
    n_levels = np.int(np.min(np.floor(np.log2(im.shape))))
    print(im.shape, n_levels, 2**n_levels)
    n_levels = 3

    sigma, mean, gain = pg_noise_bissection(im, max_err = 1E-3, n_sigmas = 3)
    aim = anscombe_transform(im, sigma, mean, gain)
    acdc, awdc = bspl_atrous(aim, n_levels, header)
    sdc = hard_threshold(awdc, n_sigmas = 5)
    ldc = label_regions(sdc)
    cdc, wdc = bspl_atrous(im, n_levels, header)
    rl = make_regions_full_props(wdc, ldc, verbose = False)
    lit = make_interscale_trees(rl, wdc, ldc, tau = 0.8, verbose = False)

    for it in lit[1:]:

        #it.plot(wdc, ldc)
        #print(it.extent)

        bspl = 1 / 16. * np.array([ 1, 4, 6, 4, 1 ])
        haar = 1 / 2. * np.array([ 1, 0, 1 ])

        #sim = im[it.bbox[0]:it.bbox[2], it.bbox[1]:it.bbox[3]]
        #scores = []
        #ssims = []
        startTime = datetime.now()
        rima = it.CG_minimization( wdc, ldc, filter = bspl, synthesis_operator = 'ADJOINT', step_size = 'FR' )
        #scores.append(20 * np.log10(  np.sum(np.sqrt(sim**2)) / ( np.sum(np.sqrt((rima - sim)**2)) ) ))
        #ssims.append( ssim(rima, sim) )
        print(datetime.now() - startTime)

        startTime = datetime.now()
        numba_rima = it.CG_minimization( wdc, ldc, filter = bspl, synthesis_operator = 'ADJOINT', step_size = 'FR' )
        #startTime = datetime.now()
        #rimb = it.CG_minimization( wdc, ldc, filter = bspl, synthesis_operator = 'SUM', step_size = 'FR' )
        #scores.append(20 * np.log10(  np.sum(np.sqrt(sim**2)) / ( np.sum(np.sqrt((rimb - sim)**2)) ) ))
        #ssims.append( ssim(rimb, sim) )

        #print(it.x_size, it.y_size, it.bbox)
        #wt, st, lt = it.tubes(wdc, ldc)
        #it.bayesian_minimization( im, wdc, ldc, bspl, verbose = True )

        #print(datetime.now() - startTime)
        #startTime = datetime.now()
        #rimc = it.SD_minimization( wdc, ldc, filter = bspl, synthesis_operator = 'ADJOINT' )
        #scores.append(20 * np.log10(  np.sum(np.sqrt(sim**2)) / ( np.sum(np.sqrt((rimc - sim)**2)) ) ))
        #ssims.append( ssim(rimc, sim) )
        #print(datetime.now() - startTime)

        #startTime = datetime.now()
        #rimd = it.SD_minimization( wdc, ldc, filter = bspl, synthesis_operator = 'SUM' )
        #scores.append(20 * np.log10(  np.sum(np.sqrt(sim**2)) / ( np.sum(np.sqrt((rimd - sim)**2)) ) ))
        #ssims.append( ssim(rimd, sim) )
        #print(datetime.now() - startTime)

        #fig, ax = plt.subplots(3, 4)

        #for i in range(4):
        #    ax[0][i].imshow(sim, norm = LogNorm())
        #    ax[0][i].set_title('sn = %2.2f\nssim = %1.2f'%(scores[i], ssims[i]))

        #ax[1][0].imshow(rima, norm = LogNorm())
        #ax[1][1].imshow(rimb, norm = LogNorm())
        #ax[1][2].imshow(rimc, norm = LogNorm())
        #ax[1][3].imshow(rimd, norm = LogNorm())

        #ax[2][0].imshow(sim - rima, norm = SymLogNorm(1E-3))
        #ax[2][1].imshow(sim - rimb, norm = SymLogNorm(1E-3))
        #ax[2][2].imshow(sim - rimc, norm = SymLogNorm(1E-3))
        #ax[2][3].imshow(sim - rimd, norm = SymLogNorm(1E-3))

        #plt.show()



from setuptools import setup

setup(name='dawis',
      version='0.0.4',
      description='Detection Algorithm with Wavelet for Intracluster light Studies',
      url='https://github.com/aellien/dawis.git',
      author='Amaël Ellien',
      author_email='a.r.j.ellien@uva.nl',
      license='GNU General Public License v3.0',
      packages=['dawis'],
      install_requires=[
          'numpy',
          'matplotlib',
          'scipy',
          'astropy',
          'scikit-image',
          'ray'
      ],
      include_package_data=True,
      package_data={'': ['gallery/*.fits']},
      zip_safe=False)



#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
#
# This file is part of DAWIS.
#
# NAME : synthesis_by_analysis.py
# AUTHOR : Amael Ellien
# LAST MODIFICATION : 07/2021
#
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
import os
import numpy as np
import logging
from astropy.io import fits
from datetime import datetime
import sys as sys
import pdb
from dawis.make_regions import *
from dawis.datacube import *
from dawis.wavelet_transforms import bspl_atrous
from dawis.hard_threshold import hard_threshold
from dawis.label_regions import label_regions
from dawis.anscombe_transforms import *
from dawis.make_regions import *
from dawis.make_interscale_trees import *
from dawis.restore_objects import *
from dawis.gif import *


def synthesis_by_analysis(indir, infile, outdir, n_cpus = 3, starting_level = 2, tau = 0.8, n_levels = None, n_sigmas = 5,\
                                gamma = 0.2, min_span = 2, max_span = 3, lvl_sep_big = 6, rm_gamma_for_big = False, monomodality = False, \
                                extent_sep = 0.1, ecc_sep = 0.95, lvl_sep_lin = 2, ceps = 1E-3, conditions = 'loop', \
                                max_iter = 500, data_dump = True, gif = True, resume = True):

    #===========================================================================
    # Check infile extension
    if infile[-5:] != '.fits':
        raise DawisWrongType('Incorrect file extension -->', infile[:-5], '.fits')

    # Logs
    outpath = os.path.join( outdir, infile[:-5] )
    namelog = ''.join(( ( outpath, '.log.txt') ))
    file_handler = logging.FileHandler( filename = namelog, mode = 'w' )
    stdout_handler = logging.StreamHandler( sys.stdout )
    logging.basicConfig( level = logging.DEBUG, handlers = [ file_handler, stdout_handler ])
    logging.info(datetime.now())
    StartTimeAlgo = datetime.now()
    logging.info('DAWIS\nInfile = %s\nInput Directory = %s\nOutput Directory = %s\nN_cpus = %d\nTau = %f\nGamma = %f\nmin_span = %d\nmax_span = %d\nConvergence Epsilon = %f'\
                                    %(infile, indir, outdir, n_cpus, tau, gamma, min_span, max_span, ceps))

    #  FITS header & image properties
    hdu = fits.open(os.path.join(indir, infile))
    im  = hdu[0].data
    header = hdu[0].header
    if not n_levels:
        n_levels = np.int(np.min(np.floor(np.log2(im.shape))))
    logging.info('Image of size %d x %d\n--> n_levels = %d, corresponding to maximum size %d\n' \
                                %(im.shape[0], im.shape[1], n_levels, 2**n_levels) )

    # Noise properties
    sigma, mean, gain = pg_noise_bissection(im, max_err = 1E-3, n_sigmas = 3)
    logging.info('Noise properties: sigma = %1.3e, mean = %1.3e, gain = %1.3e\n' %(sigma, mean, gain))

    #===========================================================================

    res = np.copy(im)
    rec = np.zeros(im.shape)
    rec_lvl = np.zeros((im.shape[0], im.shape[1], n_levels))
    cparl = []
    it = 1
    std = 1.

    for level in range( starting_level, n_levels ):

        normeps = 1.
        avnormeps = 1.
        window = []
        while avnormeps > ceps :

            start_time_it = datetime.now()
            logging.info('\n\n[ %s ] Level = %d Iteration = %d' %(datetime.now(), level, it))

            # inpaint bad reconstruction pixels and NaN pixels with noise draws
            mask = np.zeros(res.shape)
            mask[ res < -abs(mean + 5 * sigma) ] = 1.
            mask[np.where(np.isnan(res) == True)] = 1.
            draws = np.random.normal(mean, sigma, res.shape)
            mask *= draws
            res[ res < -abs(mean + 5 * sigma) ] = 0.
            res[ np.where(np.isnan(res) == True)] = 0.
            res += mask

            if ( os.path.exists(''.join(( outpath, '.ol.it%03d.pkl' %(it)))) ) & resume == True:

                logging.info('\nFound %s --> resuming iteration' %(''.join(( outpath, '.ol.it%03d.pkl' %(it)))))
                ol = read_objects_from_pickle(''.join(( outpath, '.ol.it%03d.pkl' %(it))))

            else:

                # Anscombe transform & thresholding
                aim = anscombe_transform(res, sigma, mean, gain)
                acdc, awdc = bspl_atrous(aim, level, header, conditions)
                sdc = hard_threshold(awdc, n_sigmas = n_sigmas)

                # Labels & true wavelet coefficients
                ldc = label_regions(sdc)
                cdc, wdc = bspl_atrous(res, level, header, conditions)

                # Regions of significance
                rl = make_regions_full_props(wdc, ldc, verbose = True)
                if not rl:
                    break

                # Interscale trees
                itl, ldc = make_interscale_trees(rl, wdc, ldc, tau = tau, \
                                                               min_span = min_span, \
                                                               max_span = max_span, \
                                                               lvl_sep_big = lvl_sep_big, \
                                                               monomodality = monomodality, \
                                                               n_cpus = n_cpus, \
                                                               verbose = True)
                if not itl:
                    break

                # Restoration of detected objects
                ol = restore_objects_default(itl, wdc,ldc, size_patch_small = 50, \
                                                   extent_sep = extent_sep, \
                                                   ecc_sep = ecc_sep, \
                                                   lvl_sep_lin = lvl_sep_lin, \
                                                   lvl_sep_big = lvl_sep_big, \
                                                   size_patch_big = 5, \
                                                   size_big_objects = 512, \
                                                   n_cpus = n_cpus )

            # Atom
            atom = np.zeros(res.shape)
            for object in ol:
                x_min, y_min, x_max, y_max = object.bbox
                if (rm_gamma_for_big == True) & (object.level >= lvl_sep_big):
                    atom[ x_min : x_max, y_min : y_max ] += object.image
                    rec_lvl[ x_min : x_max, y_min : y_max, object.level ] += object.image
                else:
                    atom[ x_min : x_max, y_min : y_max ] += object.image * gamma
                    rec_lvl[ x_min : x_max, y_min : y_max, object.level ] += object.image * gamma

            # Update Residuals
            res -= atom
            rec += atom

            # Convergence
            flux_res = 1 - ( np.abs(np.sum(atom) - np.sum(np.sqrt(res**2)))) / np.sum(np.sqrt(res**2))
            flux_rec = 1 - ( np.abs(np.sum(atom) - np.sum(np.sqrt(rec**2)))) / np.sum(np.sqrt(rec**2))
            logging.info('Convergence : \nres = %f rec = %f' %( flux_res, flux_rec ))
            old_std = std
            std = np.std(res)
            eps = np.sqrt( ( old_std - std )**2 ) / np.sqrt( old_std**2 )
            normeps = eps * level / len(ol)
            window.append(normeps)
            if len(window) > 5:window.pop(0)
            avnormeps = np.mean(window)
            cparl.append([ str(level), str(it), str(avnormeps), str(normeps), str(eps), str(flux_rec), str(flux_res), str(len(ol)), str( datetime.now() - start_time_it ) ])
            logging.info('Number Objects = %d, Normalized Epsilon = %f, Window Normalized Epsilon = %f' %( len(ol), normeps, avnormeps ))

            # Data Dump
            if ( os.path.exists(''.join(( outpath, '.ol.it%03d.pkl' %(it)))) ) & resume == True:
                it += 1
                continue
            else:
                write_interscale_trees_to_pickle( itl, ''.join(( outpath, '.itl.it%03d.pkl' %(it) )), overwrite = True)
                write_objects_to_pickle( ol, ''.join(( outpath, '.ol.it%03d.pkl' %(it) )), overwrite = True)
                if data_dump:
                    logging.info('Dumping data in %s' %(outdir) )
                    wdc.to_fits( ''.join(( outpath, '.wdc.it%03d.fits' %(it)) ), overwrite = True)
                    ldc.to_fits( ''.join(( outpath, '.ldc.it%03d.fits' %(it) )), overwrite = True)
                    sdc.to_fits( ''.join(( outpath, '.sdc.it%03d.fits' %(it) )), overwrite = True)
                    write_regions_to_pickle( rl, ''.join(( outpath, '.rl.it%03d.pkl' %(it) )), overwrite = True)

                if gif:
                    wdc.waveplot( name = 'Wavelet Planes\nIteration %d'%(it), show = False, save_path = ''.join(( outpath, '.wdc.it%03d.png' %(it))), origin = 'lower')
                    sdc.waveplot( name = 'Multiscale Suport\nIteration %d'%(it), show = False, save_path = ''.join(( outpath, '.sdc.it%03d.png' %(it))), origin = 'lower')
                    ldc.waveplot( name = 'Multiscale label\nIteration %d'%(it), show = False, save_path = ''.join(( outpath, '.ldc.it%03d.png' %(it))), origin = 'lower')
                    sdc.histogram_noise( name = 'Noise histogram (lvl = 0)\nIteration %d'%(it), show = False, save_path = ''.join(( outpath, '.hist.it%03d.png' %(it))))
                    plot_frame( level = level, it = it, nobj = len(ol), \
                                                        original_image = im, \
                                                        restored_image = rec, \
                                                        residuals = res, \
                                                        atom = atom, \
                                                        outpath = outpath )

            it += 1
            if it > max_iter:
                break

        if it > max_iter:
            break

    #===========================================================================
    # Write results to disk
    logging.info('Finished iterating.\nWriting results to disk in %s' %(outdir) )
    hdu_res = fits.PrimaryHDU( res, header = header )
    hdu_res.writeto( ''.join(( outpath, '.residuals.fits' )), overwrite = True )
    hdu_rec = fits.PrimaryHDU( rec, header = header )
    hdu_rec.writeto( ''.join(( outpath, '.restored.fits' )), overwrite = True )
    hdu_rec_lvl = fits.PrimaryHDU( rec_lvl, header = header )
    hdu_rec_lvl.writeto( ''.join(( outpath, '.scl.restored.fits' )), overwrite = True )

    with open( ''.join(( outpath, '.cpar.txt' )), 'w+') as conv:
        conv.write('# level, it, window_normeps, normeps, eps, n_obj, time_it\n')
        for l in cparl:
            conv.write('%s\n' %(' '.join(l)))

    plot_convergence( cparl = cparl, outpath = outpath, marker = 'o', linewidth = 2 )

    if gif:
        make_gif( framerate = 3, outpath = outpath )

    logging.info( datetime.now() - StartTimeAlgo )


def load_iteration( it, outpath ):
    '''doc todo'''

    wdc = wavelet_datacube.from_fits(''.join(( outpath, '.wdc.it%03d.fits' %(it)) ))
    ldc = label_datacube.from_fits( ''.join(( outpath, '.ldc.it%03d.fits' %(it) )))
    rl = read_regions_from_pickle( ''.join(( outpath, '.rl.it%03d.pkl' %(it) )))
    itl = read_interscale_trees_from_pickle( ''.join(( outpath, '.itl.it%03d.pkl' %(it) )))
    ol = read_objects_from_pickle( ''.join(( outpath, '.ol.it%03d.pkl' %(it) )))
    return wdc, ldc, rl, itl, ol

if __name__=="__main__":

    if len(sys.argv) < 4:
        raise SyntaxError("Insufficient arguments.")

    main(sys.argv[1], sys.argv[2], sys.argv[3], *sys.argv[4:])
